from operator import add
from operator import mul
from operator import sub

import numpy as np

from ..base import EphemeralNode
from ..base import FunctionalNode
from ..base import TerminalNode
from ..base import UniversalSet
from ..base import create_operator
from ..base._tree import save_div
from ..optimizers import DifferentialEvolution
from ..optimizers import GeneticAlgorithm
from ..optimizers import GeneticProgramming
from ..optimizers import SHADE
from ..optimizers import SHAGA
from ..optimizers import SelfCGA
from ..optimizers import SelfCGP
from ..optimizers import jDE
from ..utils._metrics import coefficient_determination
from thefittest.utils.random import generator1
from thefittest.utils.random import generator2


def test_GeneticAlgorithm_start_settings():
    def fitness_function(x):
        return np.sum(x, axis=1, dtype=np.float64)

    iters = 100
    pop_size = 50
    str_len = 200

    # simple start
    optimizer = GeneticAlgorithm(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        str_len=str_len,
        optimal_value=None,
        termination_error_value=0,
        no_increase_num=None,
        minimization=True,
        show_progress_each=1,
        keep_history=True,
    )

    optimizer.fit()

    fittest = optimizer.get_fittest()
    assert isinstance(fittest, dict)

    stats = optimizer.get_stats()

    assert optimizer.get_remains_calls() == 0
    assert len(stats["max_fitness"]) == iters
    assert optimizer._sign == -1

    # start with the no_increase_num is equal 15
    def fitness_function(x):
        return np.ones(len(x), dtype=np.float64)

    no_increase_num = 15
    optimizer = GeneticAlgorithm(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        str_len=str_len,
        optimal_value=None,
        termination_error_value=0,
        no_increase_num=no_increase_num,
        minimization=False,
        show_progress_each=1,
        keep_history=True,
    )

    optimizer.fit()

    assert optimizer.get_remains_calls() == pop_size * (iters - no_increase_num - 1)
    assert optimizer._sign == 1

    # start with the optimal_value is equal 1
    optimizer = GeneticAlgorithm(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        str_len=str_len,
        optimal_value=1,
        termination_error_value=0,
        no_increase_num=None,
        minimization=False,
        show_progress_each=1,
        keep_history=True,
    )

    optimizer.fit()
    assert optimizer.get_remains_calls() == pop_size * (iters - 1)


def test_GeneticAlgorithm_set_strategy():
    def fitness_function(x):
        return np.sum(x, axis=1, dtype=np.float64)

    iters = 10
    pop_size = 8
    str_len = 30

    selections = (
        "proportional",
        "rank",
        "tournament_k",
        "tournament_3",
        "tournament_5",
        "tournament_7",
    )
    crossover = (
        "empty",
        "one_point",
        "two_point",
        "uniform_2",
        "uniform_7",
        "uniform_k",
        "uniform_prop_2",
        "uniform_prop_7",
        "uniform_prop_k",
        "uniform_rank_2",
        "uniform_rank_7",
        "uniform_rank_k",
        "uniform_tour_3",
        "uniform_tour_7",
        "uniform_tour_k",
    )

    mutation = ("weak", "average", "strong", "custom_rate")

    tour_size = (3, 4, 5)
    parents_num = (7, 8, 9)
    mutation_rate = (0.11, 0.22)

    for selections_i in selections:
        for mutation_rate_i in mutation_rate:
            for crossover_i in crossover:
                for mutation_i in mutation:
                    for tour_size_i in tour_size:
                        for parents_num_i in parents_num:
                            random_state = np.random.randint(0, 100)
                            initial_population = GeneticAlgorithm.binary_string_population(
                                pop_size=pop_size, str_len=str_len
                            )
                            optimizer = GeneticAlgorithm(
                                fitness_function,
                                iters=iters,
                                pop_size=pop_size,
                                str_len=str_len,
                                optimal_value=None,
                                termination_error_value=0,
                                no_increase_num=None,
                                minimization=False,
                                show_progress_each=None,
                                keep_history=True,
                                init_population=initial_population,
                                selection=selections_i,
                                crossover=crossover_i,
                                mutation=mutation_i,
                                tour_size=tour_size_i,
                                elitism=True,
                                parents_num=parents_num_i,
                                mutation_rate=mutation_rate_i,
                                random_state=random_state,
                            )

                            assert optimizer._specified_selection == selections_i
                            assert optimizer._specified_crossover == crossover_i
                            assert optimizer._specified_mutation == mutation_i
                            assert optimizer._tour_size == tour_size_i
                            assert optimizer._parents_num == parents_num_i
                            assert optimizer._mutation_rate == mutation_rate_i

                            if selections_i[-2:] == "_k":
                                assert (
                                    optimizer._selection_pool[optimizer._specified_selection][1]
                                    == tour_size_i
                                )
                            if crossover_i[-2:] == "_k":
                                assert (
                                    optimizer._crossover_pool[optimizer._specified_crossover][1]
                                    == parents_num_i
                                )
                            if mutation_i == "custom_rate":
                                assert (
                                    optimizer._mutation_pool[optimizer._specified_mutation][1]
                                    == mutation_rate_i
                                )

                            optimizer.fit()

                            stats = optimizer.get_stats()
                            assert np.all(stats["population_g"][0] == initial_population)

                            genotype_1 = optimizer.get_fittest()["genotype"]

                            optimizer = GeneticAlgorithm(
                                fitness_function,
                                iters=iters,
                                pop_size=pop_size,
                                str_len=str_len,
                                optimal_value=None,
                                termination_error_value=0,
                                no_increase_num=None,
                                minimization=False,
                                show_progress_each=None,
                                keep_history=True,
                                init_population=initial_population,
                                selection=selections_i,
                                crossover=crossover_i,
                                mutation=mutation_i,
                                tour_size=tour_size_i,
                                elitism=True,
                                parents_num=parents_num_i,
                                mutation_rate=mutation_rate_i,
                                random_state=random_state,
                            )

                            optimizer.fit()

                            genotype_2 = optimizer.get_fittest()["genotype"]
                            assert np.all(genotype_1 == genotype_2)


def test_SelfCGA_start_settings():
    def fitness_function(x):
        return np.sum(x, axis=1, dtype=np.float64)

    iters = 100
    pop_size = 50
    str_len = 200

    # simple start
    optimizer = SelfCGA(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        str_len=str_len,
        optimal_value=None,
        termination_error_value=0,
        no_increase_num=None,
        minimization=True,
        show_progress_each=1,
        keep_history=True,
    )

    optimizer.fit()

    fittest = optimizer.get_fittest()
    assert isinstance(fittest, dict)

    stats = optimizer.get_stats()

    assert optimizer.get_remains_calls() == 0
    assert len(stats["max_fitness"]) == iters
    assert optimizer._sign == -1

    # start with the no_increase_num is equal 15
    def fitness_function(x):
        return np.ones(len(x), dtype=np.float64)

    no_increase_num = 15
    optimizer = SelfCGA(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        str_len=str_len,
        optimal_value=None,
        termination_error_value=0,
        no_increase_num=no_increase_num,
        minimization=False,
        show_progress_each=1,
        keep_history=True,
    )

    optimizer.fit()

    assert optimizer.get_remains_calls() == pop_size * (iters - no_increase_num - 1)
    assert optimizer._sign == 1

    # start with the optimal_value is equal 1
    optimizer = SelfCGA(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        str_len=str_len,
        optimal_value=1,
        termination_error_value=0,
        no_increase_num=None,
        minimization=False,
        show_progress_each=1,
        keep_history=True,
    )

    optimizer.fit()
    assert optimizer.get_remains_calls() == pop_size * (iters - 1)


def test_SelfCGA_set_strategy():
    def fitness_function(x):
        return np.sum(x, axis=1, dtype=np.float64)

    iters = 10
    pop_size = 10
    str_len = 5
    initial_population = SelfCGA.binary_string_population(pop_size=pop_size, str_len=str_len)
    tour_size = 8
    parents_num = 8
    mutation_rate = 0.5

    optimizer = SelfCGA(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        str_len=str_len,
        optimal_value=None,
        termination_error_value=0,
        no_increase_num=None,
        minimization=False,
        show_progress_each=None,
        keep_history=True,
        init_population=initial_population,
        selections=(
            "proportional",
            "rank",
            "tournament_k",
            "tournament_3",
        ),
        crossovers=(
            "empty",
            "one_point",
            "two_point",
            "uniform_2",
            "uniform_k",
            "uniform_prop_2",
            "uniform_prop_k",
            "uniform_rank_2",
            "uniform_rank_k",
            "uniform_tour_3",
            "uniform_tour_k",
        ),
        mutations=("weak", "custom_rate"),
        tour_size=tour_size,
        elitism=True,
        parents_num=parents_num,
        mutation_rate=mutation_rate,
        random_state=42,
    )
    optimizer.fit()
    stats = optimizer.get_stats()
    assert np.all(stats["population_g"][0] == initial_population)

    genotype_1 = optimizer.get_fittest()["genotype"]

    optimizer = SelfCGA(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        str_len=str_len,
        optimal_value=None,
        termination_error_value=0,
        no_increase_num=None,
        minimization=False,
        show_progress_each=None,
        keep_history=True,
        init_population=initial_population,
        selections=(
            "proportional",
            "rank",
            "tournament_k",
            "tournament_3",
        ),
        crossovers=(
            "empty",
            "one_point",
            "two_point",
            "uniform_2",
            "uniform_k",
            "uniform_prop_2",
            "uniform_prop_k",
            "uniform_rank_2",
            "uniform_rank_k",
            "uniform_tour_3",
            "uniform_tour_k",
        ),
        mutations=("weak", "custom_rate"),
        tour_size=tour_size,
        elitism=True,
        parents_num=parents_num,
        mutation_rate=mutation_rate,
        random_state=42,
    )

    optimizer.fit()

    genotype_2 = optimizer.get_fittest()["genotype"]

    assert np.all(genotype_1 == genotype_2)


def test_GeneticProgramming_start_settings():

    def problem(x):
        return 3 * x[:, 0] ** 2 + 2 * x[:, 0] + 5

    function = problem
    left_border = -4.5
    right_border = 4.5
    sample_size = 300
    n_dimension = 1

    iters = 20
    pop_size = 15

    X = np.array(
        [np.linspace(left_border, right_border, sample_size) for _ in range(n_dimension)]
    ).T
    y = function(X)

    functional_set = (
        FunctionalNode(create_operator("({} + {})", "add", "+", add)),
        FunctionalNode(create_operator("({} * {})", "mul", "*", mul)),
        FunctionalNode(create_operator("({} - {})", "sub", "-", sub)),
        FunctionalNode(create_operator("({} / {})", "div", "/", save_div)),
    )

    terminal_set = [TerminalNode(X[:, i], f"x{i}") for i in range(n_dimension)]
    terminal_set.extend([EphemeralNode(generator1), EphemeralNode(generator2)])
    uniset = UniversalSet(functional_set, terminal_set)

    def fitness_function(trees):
        fitness = []
        for tree in trees:
            y_pred = tree() * np.ones(len(y))
            fitness.append(-coefficient_determination(y, y_pred))
        return np.array(fitness)

    # simple start
    optimizer = GeneticProgramming(
        fitness_function=fitness_function,
        uniset=uniset,
        pop_size=pop_size,
        iters=iters,
        optimal_value=None,
        termination_error_value=0,
        no_increase_num=None,
        show_progress_each=1,
        minimization=True,
        keep_history=True,
    )

    optimizer.fit()

    fittest = optimizer.get_fittest()
    assert isinstance(fittest, dict)

    stats = optimizer.get_stats()

    assert optimizer.get_remains_calls() == 0
    assert len(stats["max_fitness"]) == iters
    assert optimizer._sign == -1

    # start with the no_increase_num is equal 15
    def fitness_function(x):
        return np.ones(len(x), dtype=np.float64)

    no_increase_num = 15

    optimizer = GeneticProgramming(
        fitness_function=fitness_function,
        uniset=uniset,
        pop_size=pop_size,
        iters=iters,
        optimal_value=None,
        termination_error_value=0,
        no_increase_num=no_increase_num,
        show_progress_each=1,
        minimization=False,
        keep_history=True,
    )

    optimizer.fit()

    assert optimizer.get_remains_calls() == pop_size * (iters - no_increase_num - 1)
    assert optimizer._sign == 1

    # start with the optimal_value is equal 1
    optimizer = GeneticProgramming(
        fitness_function=fitness_function,
        uniset=uniset,
        pop_size=pop_size,
        iters=iters,
        optimal_value=1,
        termination_error_value=0,
        no_increase_num=None,
        show_progress_each=1,
        minimization=False,
        keep_history=True,
    )

    optimizer.fit()
    assert optimizer.get_remains_calls() == pop_size * (iters - 1)


def test_GeneticProgramming_set_strategy():

    def problem(x):
        return 3 * x[:, 0] ** 2 + 2 * x[:, 0] + 5

    function = problem
    left_border = -4.5
    right_border = 4.5
    sample_size = 30
    n_dimension = 1

    iters = 2
    pop_size = 10

    X = np.array(
        [np.linspace(left_border, right_border, sample_size) for _ in range(n_dimension)]
    ).T
    y = function(X)

    functional_set = (
        FunctionalNode(create_operator("({} + {})", "add", "+", add)),
        FunctionalNode(create_operator("({} * {})", "mul", "*", mul)),
        FunctionalNode(create_operator("({} - {})", "sub", "-", sub)),
        FunctionalNode(create_operator("({} / {})", "div", "/", save_div)),
    )

    terminal_set = [TerminalNode(X[:, i], f"x{i}") for i in range(n_dimension)]
    terminal_set.extend([EphemeralNode(generator1), EphemeralNode(generator2)])
    uniset = UniversalSet(functional_set, terminal_set)

    def fitness_function(trees):
        fitness = []
        for tree in trees:
            y_pred = tree() * np.ones(len(y))
            fitness.append(-coefficient_determination(y, y_pred))
        return np.array(fitness)

    initial_population = GeneticProgramming.half_and_half(
        pop_size=pop_size, uniset=uniset, max_level=14
    )
    selections = (
        "proportional",
        "rank",
        "tournament_k",
        "tournament_3",
    )
    crossover = (
        "gp_empty",
        "gp_standard",
        "gp_one_point",
        "gp_uniform_2",
        "gp_uniform_k",
        "gp_uniform_prop_2",
        "gp_uniform_prop_k",
        "gp_uniform_rank_2",
        "gp_uniform_rank_k",
        "gp_uniform_tour_3",
        "gp_uniform_tour_k",
    )

    mutation = (
        "gp_weak_point",
        "gp_custom_rate_point",
        "gp_weak_grow",
        "gp_custom_rate_grow",
        "gp_weak_swap",
        "gp_custom_rate_swap",
        "gp_weak_shrink",
        "gp_custom_rate_shrink",
    )

    for selections_i in selections:
        for crossover_i in crossover:
            for mutation_i in mutation:
                initial_population = GeneticProgramming.half_and_half(
                    pop_size=pop_size, uniset=uniset, max_level=14
                )
                random_state = np.random.randint(0, 100)
                optimizer = GeneticProgramming(
                    fitness_function,
                    iters=iters,
                    pop_size=pop_size,
                    uniset=uniset,
                    optimal_value=None,
                    termination_error_value=0,
                    no_increase_num=None,
                    minimization=False,
                    show_progress_each=None,
                    keep_history=True,
                    init_population=initial_population,
                    selection=selections_i,
                    crossover=crossover_i,
                    mutation=mutation_i,
                    tour_size=3,
                    elitism=True,
                    parents_num=3,
                    mutation_rate=0.33,
                    random_state=random_state,
                )

                assert optimizer._specified_selection == selections_i
                assert optimizer._specified_crossover == crossover_i
                assert optimizer._specified_mutation == mutation_i
                assert optimizer._tour_size == 3
                assert optimizer._parents_num == 3
                assert optimizer._mutation_rate == 0.33

                if selections_i[-2:] == "_k":
                    assert optimizer._selection_pool[optimizer._specified_selection][1] == 3
                if crossover_i[-2:] == "_k":
                    assert optimizer._crossover_pool[optimizer._specified_crossover][1] == 3
                if mutation_i[:9] == "gp_custom":
                    assert optimizer._mutation_pool[optimizer._specified_mutation][1] == 0.33

                optimizer.fit()

                stats = optimizer.get_stats()
                assert np.all(stats["population_g"][0] == initial_population)
                genotype_1 = optimizer.get_fittest()["genotype"]

                optimizer = GeneticProgramming(
                    fitness_function,
                    iters=iters,
                    pop_size=pop_size,
                    uniset=uniset,
                    optimal_value=None,
                    termination_error_value=0,
                    no_increase_num=None,
                    minimization=False,
                    show_progress_each=None,
                    keep_history=True,
                    init_population=initial_population,
                    selection=selections_i,
                    crossover=crossover_i,
                    mutation=mutation_i,
                    tour_size=3,
                    elitism=True,
                    parents_num=3,
                    mutation_rate=0.33,
                    random_state=random_state,
                )
                optimizer.fit()

                genotype_2 = optimizer.get_fittest()["genotype"]
                print(mutation_i, crossover_i, selections_i)
                assert np.all(genotype_1 == genotype_2)


def test_SelfCGP_start_settings():

    def problem(x):
        return 3 * x[:, 0] ** 2 + 2 * x[:, 0] + 5

    function = problem
    left_border = -4.5
    right_border = 4.5
    sample_size = 300
    n_dimension = 1

    iters = 20
    pop_size = 15

    X = np.array(
        [np.linspace(left_border, right_border, sample_size) for _ in range(n_dimension)]
    ).T
    y = function(X)

    functional_set = (
        FunctionalNode(create_operator("({} + {})", "add", "+", add)),
        FunctionalNode(create_operator("({} * {})", "mul", "*", mul)),
        FunctionalNode(create_operator("({} - {})", "sub", "-", sub)),
        FunctionalNode(create_operator("({} / {})", "div", "/", save_div)),
    )

    terminal_set = [TerminalNode(X[:, i], f"x{i}") for i in range(n_dimension)]
    terminal_set.extend([EphemeralNode(generator1), EphemeralNode(generator2)])
    uniset = UniversalSet(functional_set, terminal_set)

    def fitness_function(trees):
        fitness = []
        for tree in trees:
            y_pred = tree() * np.ones(len(y))
            fitness.append(-coefficient_determination(y, y_pred))
        return np.array(fitness)

    # simple start
    optimizer = SelfCGP(
        fitness_function=fitness_function,
        uniset=uniset,
        pop_size=pop_size,
        iters=iters,
        optimal_value=None,
        termination_error_value=0,
        no_increase_num=None,
        show_progress_each=1,
        minimization=True,
        keep_history=True,
    )

    optimizer.fit()

    fittest = optimizer.get_fittest()
    assert isinstance(fittest, dict)

    stats = optimizer.get_stats()

    assert optimizer.get_remains_calls() == 0
    assert len(stats["max_fitness"]) == iters
    assert optimizer._sign == -1

    # start with the no_increase_num is equal 15
    def fitness_function(x):
        return np.ones(len(x), dtype=np.float64)

    no_increase_num = 15

    optimizer = SelfCGP(
        fitness_function=fitness_function,
        uniset=uniset,
        pop_size=pop_size,
        iters=iters,
        optimal_value=None,
        termination_error_value=0,
        no_increase_num=no_increase_num,
        show_progress_each=1,
        minimization=False,
        keep_history=True,
    )

    optimizer.fit()

    assert optimizer.get_remains_calls() == pop_size * (iters - no_increase_num - 1)
    assert optimizer._sign == 1

    # start with the optimal_value is equal 1
    optimizer = SelfCGP(
        fitness_function=fitness_function,
        uniset=uniset,
        pop_size=pop_size,
        iters=iters,
        optimal_value=1,
        termination_error_value=0,
        no_increase_num=None,
        show_progress_each=1,
        minimization=False,
        keep_history=True,
    )

    optimizer.fit()
    assert optimizer.get_remains_calls() == pop_size * (iters - 1)


def test_SelfCGP_set_strategy():

    def problem(x):
        return 3 * x[:, 0] ** 2 + 2 * x[:, 0] + 5

    function = problem
    left_border = -4.5
    right_border = 4.5
    sample_size = 300
    n_dimension = 1

    X = np.array(
        [np.linspace(left_border, right_border, sample_size) for _ in range(n_dimension)]
    ).T
    y = function(X)

    functional_set = (
        FunctionalNode(create_operator("({} + {})", "add", "+", add)),
        FunctionalNode(create_operator("({} * {})", "mul", "*", mul)),
        FunctionalNode(create_operator("({} - {})", "sub", "-", sub)),
        FunctionalNode(create_operator("({} / {})", "div", "/", save_div)),
    )

    terminal_set = [TerminalNode(X[:, i], f"x{i}") for i in range(n_dimension)]
    terminal_set.extend([EphemeralNode(generator1), EphemeralNode(generator2)])
    uniset = UniversalSet(functional_set, terminal_set)

    def fitness_function(trees):
        fitness = []
        for tree in trees:
            y_pred = tree() * np.ones(len(y))
            fitness.append(-coefficient_determination(y, y_pred))
        return np.array(fitness)

    iters = 10
    pop_size = 10
    initial_population = SelfCGP.half_and_half(pop_size=pop_size, uniset=uniset, max_level=14)
    tour_size = 8
    parents_num = 8
    mutation_rate = 0.5

    optimizer = SelfCGP(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        uniset=uniset,
        optimal_value=None,
        termination_error_value=0,
        no_increase_num=None,
        minimization=False,
        show_progress_each=None,
        keep_history=True,
        init_population=initial_population,
        selections=(
            "proportional",
            "rank",
            "tournament_k",
            "tournament_3",
            "tournament_5",
            "tournament_7",
        ),
        crossovers=(
            "gp_empty",
            "gp_standard",
            "gp_one_point",
            "gp_uniform_2",
            "gp_uniform_7",
            "gp_uniform_k",
            "gp_uniform_prop_2",
            "gp_uniform_prop_7",
            "gp_uniform_prop_k",
            "gp_uniform_rank_2",
            "gp_uniform_rank_7",
            "gp_uniform_rank_k",
            "gp_uniform_tour_3",
            "gp_uniform_tour_7",
            "gp_uniform_tour_k",
        ),
        mutations=(
            "gp_weak_point",
            "gp_average_point",
            "gp_strong_point",
            "gp_custom_rate_point",
            "gp_weak_grow",
            "gp_average_grow",
            "gp_strong_grow",
            "gp_custom_rate_grow",
            "gp_weak_swap",
            "gp_average_swap",
            "gp_strong_swap",
            "gp_custom_rate_swap",
            "gp_weak_shrink",
            "gp_average_shrink",
            "gp_strong_shrink",
            "gp_custom_rate_shrink",
        ),
        tour_size=tour_size,
        elitism=True,
        parents_num=parents_num,
        mutation_rate=mutation_rate,
        random_state=42,
    )
    optimizer.fit()
    stats = optimizer.get_stats()
    assert np.all(stats["population_g"][0] == initial_population)
    genotype_1 = optimizer.get_fittest()["genotype"]

    optimizer = SelfCGP(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        uniset=uniset,
        optimal_value=None,
        termination_error_value=0,
        no_increase_num=None,
        minimization=False,
        show_progress_each=None,
        keep_history=True,
        init_population=initial_population,
        selections=(
            "proportional",
            "rank",
            "tournament_k",
            "tournament_3",
            "tournament_5",
            "tournament_7",
        ),
        crossovers=(
            "gp_empty",
            "gp_standard",
            "gp_one_point",
            "gp_uniform_2",
            "gp_uniform_7",
            "gp_uniform_k",
            "gp_uniform_prop_2",
            "gp_uniform_prop_7",
            "gp_uniform_prop_k",
            "gp_uniform_rank_2",
            "gp_uniform_rank_7",
            "gp_uniform_rank_k",
            "gp_uniform_tour_3",
            "gp_uniform_tour_7",
            "gp_uniform_tour_k",
        ),
        mutations=(
            "gp_weak_point",
            "gp_average_point",
            "gp_strong_point",
            "gp_custom_rate_point",
            "gp_weak_grow",
            "gp_average_grow",
            "gp_strong_grow",
            "gp_custom_rate_grow",
            "gp_weak_swap",
            "gp_average_swap",
            "gp_strong_swap",
            "gp_custom_rate_swap",
            "gp_weak_shrink",
            "gp_average_shrink",
            "gp_strong_shrink",
            "gp_custom_rate_shrink",
        ),
        tour_size=tour_size,
        elitism=True,
        parents_num=parents_num,
        mutation_rate=mutation_rate,
        random_state=42,
    )

    optimizer.fit()

    genotype_2 = optimizer.get_fittest()["genotype"]

    assert np.all(genotype_1 == genotype_2)


def test_DifferentialEvolution_start_settings():
    def fitness_function(x):
        return np.sum(x, axis=1, dtype=np.float64)

    iters = 100
    pop_size = 50
    n_vars = 10
    left = np.full(n_vars, -1, dtype=np.float64)
    right = np.full(n_vars, 1, dtype=np.float64)

    # simple start
    optimizer = DifferentialEvolution(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        left_border=left,
        right_border=right,
        num_variables=n_vars,
        optimal_value=None,
        termination_error_value=0,
        no_increase_num=None,
        minimization=True,
        show_progress_each=1,
        keep_history=True,
    )

    optimizer.fit()

    fittest = optimizer.get_fittest()
    assert isinstance(fittest, dict)

    stats = optimizer.get_stats()

    assert optimizer.get_remains_calls() == 0
    assert len(stats["max_fitness"]) == iters
    assert optimizer._sign == -1

    # start with the no_increase_num is equal 15
    def fitness_function(x):
        return np.ones(len(x), dtype=np.float64)

    no_increase_num = 15
    optimizer = DifferentialEvolution(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        left_border=-1,
        right_border=1,
        num_variables=n_vars,
        optimal_value=None,
        termination_error_value=0,
        no_increase_num=no_increase_num,
        minimization=False,
        show_progress_each=1,
        keep_history=True,
    )

    optimizer.fit()

    assert optimizer.get_remains_calls() == pop_size * (iters - no_increase_num - 1)
    assert optimizer._sign == 1

    # start with the optimal_value is equal 1
    optimizer = DifferentialEvolution(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        left_border=left,
        right_border=right,
        num_variables=n_vars,
        optimal_value=1,
        termination_error_value=0,
        no_increase_num=None,
        minimization=False,
        show_progress_each=1,
        keep_history=True,
    )

    optimizer.fit()
    assert optimizer.get_remains_calls() == pop_size * (iters - 1)


def test_DifferentialEvolution_set_strategy():

    def fitness_function(x):
        return np.sum(x, axis=1, dtype=np.float64)

    iters = 2
    pop_size = 10
    n_vars = 10
    left = np.full(n_vars, -1, dtype=np.float64)
    right = np.full(n_vars, 1, dtype=np.float64)

    mutations = ("best_1", "rand_1", "current_to_best_1", "rand_to_best1", "best_2", "rand_2")
    F = (0.5, 0.3)
    CR = (0.31, 0.33)

    for mutation_i in mutations:
        for F_i in F:
            for CR_i in CR:
                initial_population = DifferentialEvolution.float_population(
                    pop_size=pop_size, left_border=left, right_border=right, num_variables=n_vars
                )
                random_state = np.random.randint(0, 100)
                optimizer = DifferentialEvolution(
                    fitness_function,
                    iters=iters,
                    pop_size=pop_size,
                    left_border=left,
                    right_border=right,
                    num_variables=n_vars,
                    optimal_value=None,
                    termination_error_value=0,
                    no_increase_num=None,
                    minimization=True,
                    show_progress_each=1,
                    keep_history=True,
                    mutation=mutation_i,
                    init_population=initial_population,
                    F=F_i,
                    CR=CR_i,
                    random_state=random_state,
                )
                optimizer.fit()
                stats = optimizer.get_stats()

                assert optimizer._specified_mutation == mutation_i
                assert optimizer._F == F_i
                assert optimizer._CR == CR_i

                assert np.all(stats["population_g"][0] == initial_population)

                genotype_1 = optimizer.get_fittest()["genotype"]

                optimizer = DifferentialEvolution(
                    fitness_function,
                    iters=iters,
                    pop_size=pop_size,
                    left_border=left,
                    right_border=right,
                    num_variables=n_vars,
                    optimal_value=None,
                    termination_error_value=0,
                    no_increase_num=None,
                    minimization=True,
                    show_progress_each=1,
                    keep_history=True,
                    mutation=mutation_i,
                    init_population=initial_population,
                    F=F_i,
                    CR=CR_i,
                    random_state=random_state,
                )

                optimizer.fit()

                genotype_2 = optimizer.get_fittest()["genotype"]

                assert np.all(genotype_1 == genotype_2)


def test_jDE_start_settings():
    def fitness_function(x):
        return np.sum(x, axis=1, dtype=np.float64)

    iters = 100
    pop_size = 50
    n_vars = 10
    left = np.full(n_vars, -1, dtype=np.float64)
    right = np.full(n_vars, 1, dtype=np.float64)

    # simple start
    optimizer = jDE(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        left_border=left,
        right_border=right,
        num_variables=n_vars,
        optimal_value=None,
        termination_error_value=0,
        no_increase_num=None,
        minimization=True,
        show_progress_each=1,
        keep_history=True,
    )

    optimizer.fit()

    fittest = optimizer.get_fittest()
    assert isinstance(fittest, dict)

    stats = optimizer.get_stats()

    assert optimizer.get_remains_calls() == 0
    assert len(stats["max_fitness"]) == iters
    assert optimizer._sign == -1

    # start with the no_increase_num is equal 15
    def fitness_function(x):
        return np.ones(len(x), dtype=np.float64)

    no_increase_num = 15
    optimizer = jDE(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        left_border=-1,
        right_border=1,
        num_variables=n_vars,
        optimal_value=None,
        termination_error_value=0,
        no_increase_num=no_increase_num,
        minimization=False,
        show_progress_each=1,
        keep_history=True,
    )

    optimizer.fit()

    assert optimizer.get_remains_calls() == pop_size * (iters - no_increase_num - 1)
    assert optimizer._sign == 1

    # start with the optimal_value is equal 1
    optimizer = jDE(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        left_border=left,
        right_border=right,
        num_variables=n_vars,
        optimal_value=1,
        termination_error_value=0,
        no_increase_num=None,
        minimization=False,
        show_progress_each=1,
        keep_history=True,
    )

    optimizer.fit()
    assert optimizer.get_remains_calls() == pop_size * (iters - 1)


def test_jDE_set_strategy():
    def fitness_function(x):
        return np.sum(x, axis=1, dtype=np.float64)

    iters = 2
    pop_size = 10
    n_vars = 10
    left = np.full(n_vars, -1, dtype=np.float64)
    right = np.full(n_vars, 1, dtype=np.float64)

    mutations = ("best_1", "rand_1", "current_to_best_1", "rand_to_best1", "best_2", "rand_2")
    F_min = (0.05, 0.1)
    F_max = (0.9, 0.95)
    t_F = (0.051, 0.21)
    t_CR = (0.05, 0.1)

    for mutation_i in mutations:
        for F_min_i in F_min:
            for F_max_i in F_max:
                for t_F_i in t_F:
                    for t_CR_i in t_CR:
                        initial_population = jDE.float_population(
                            pop_size=pop_size,
                            left_border=left,
                            right_border=right,
                            num_variables=n_vars,
                        )
                        random_state = np.random.randint(0, 100)
                        optimizer = jDE(
                            fitness_function,
                            iters=iters,
                            pop_size=pop_size,
                            left_border=left,
                            right_border=right,
                            num_variables=n_vars,
                            optimal_value=None,
                            termination_error_value=0,
                            no_increase_num=None,
                            minimization=True,
                            show_progress_each=1,
                            keep_history=True,
                            mutation=mutation_i,
                            F_min=F_min_i,
                            F_max=F_max_i,
                            t_CR=t_CR_i,
                            t_F=t_F_i,
                            init_population=initial_population,
                            random_state=random_state,
                        )
                        optimizer.fit()
                        stats = optimizer.get_stats()

                        assert optimizer._specified_mutation == mutation_i
                        assert optimizer._F_min == F_min_i
                        assert optimizer._F_max == F_max_i
                        assert optimizer._t_CR == t_CR_i
                        assert optimizer._t_F == t_F_i
                        print(stats.keys())
                        assert np.all(stats["population_g"][0] == initial_population)

                        genotype_1 = optimizer.get_fittest()["genotype"]

                        optimizer = jDE(
                            fitness_function,
                            iters=iters,
                            pop_size=pop_size,
                            left_border=left,
                            right_border=right,
                            num_variables=n_vars,
                            optimal_value=None,
                            termination_error_value=0,
                            no_increase_num=None,
                            minimization=True,
                            show_progress_each=1,
                            keep_history=True,
                            mutation=mutation_i,
                            F_min=F_min_i,
                            F_max=F_max_i,
                            t_CR=t_CR_i,
                            t_F=t_F_i,
                            init_population=initial_population,
                            random_state=random_state,
                        )

                        optimizer.fit()

                        genotype_2 = optimizer.get_fittest()["genotype"]

                        assert np.all(genotype_1 == genotype_2)


def test_SHADE_start_settings():
    def fitness_function(x):
        return np.sum(x, axis=1, dtype=np.float64)

    iters = 100
    pop_size = 50
    n_vars = 10
    left = np.full(n_vars, -1, dtype=np.float64)
    right = np.full(n_vars, 1, dtype=np.float64)

    # simple start
    optimizer = SHADE(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        left_border=left,
        right_border=right,
        num_variables=n_vars,
        optimal_value=None,
        termination_error_value=0,
        no_increase_num=None,
        minimization=True,
        show_progress_each=1,
        keep_history=True,
        random_state=42,
    )

    optimizer.fit()

    genotype_1 = optimizer.get_fittest()["genotype"]

    optimizer = SHADE(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        left_border=-1,
        right_border=1,
        num_variables=n_vars,
        optimal_value=None,
        termination_error_value=0,
        no_increase_num=None,
        minimization=True,
        show_progress_each=1,
        keep_history=True,
        random_state=42,
    )
    optimizer.fit()

    genotype_2 = optimizer.get_fittest()["genotype"]

    assert np.all(genotype_1 == genotype_2)

    fittest = optimizer.get_fittest()
    assert isinstance(fittest, dict)

    stats = optimizer.get_stats()

    assert optimizer.get_remains_calls() == 0
    assert len(stats["max_fitness"]) == iters
    assert optimizer._sign == -1

    # start with the no_increase_num is equal 15
    def fitness_function(x):
        return np.ones(len(x), dtype=np.float64)

    no_increase_num = 15
    optimizer = SHADE(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        left_border=left,
        right_border=right,
        num_variables=n_vars,
        optimal_value=None,
        termination_error_value=0,
        no_increase_num=no_increase_num,
        minimization=False,
        show_progress_each=1,
        keep_history=True,
    )

    optimizer.fit()

    assert optimizer.get_remains_calls() == pop_size * (iters - no_increase_num - 1)
    assert optimizer._sign == 1

    # start with the optimal_value is equal 1
    optimizer = SHADE(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        left_border=left,
        right_border=right,
        num_variables=n_vars,
        optimal_value=1,
        termination_error_value=0,
        no_increase_num=None,
        minimization=False,
        show_progress_each=1,
        keep_history=True,
    )

    optimizer.fit()
    assert optimizer.get_remains_calls() == pop_size * (iters - 1)


def test_SHAGA_start_settings():
    def fitness_function(x):
        return np.sum(x, axis=1, dtype=np.float64)

    iters = 100
    pop_size = 50
    str_len = 200

    # simple start
    optimizer = SHAGA(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        str_len=str_len,
        optimal_value=None,
        termination_error_value=0,
        no_increase_num=None,
        minimization=True,
        show_progress_each=1,
        keep_history=True,
        random_state=42,
    )

    optimizer.fit()

    genotype_1 = optimizer.get_fittest()["genotype"]

    optimizer = SHAGA(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        str_len=str_len,
        optimal_value=None,
        termination_error_value=0,
        no_increase_num=None,
        minimization=True,
        show_progress_each=1,
        keep_history=True,
        random_state=42,
    )

    optimizer.fit()

    genotype_2 = optimizer.get_fittest()["genotype"]

    assert np.all(genotype_1 == genotype_2)

    fittest = optimizer.get_fittest()
    assert isinstance(fittest, dict)

    stats = optimizer.get_stats()

    assert optimizer.get_remains_calls() == 0
    assert len(stats["max_fitness"]) == iters
    assert optimizer._sign == -1

    # start with the no_increase_num is equal 15
    def fitness_function(x):
        return np.ones(len(x), dtype=np.float64)

    no_increase_num = 15
    optimizer = SHAGA(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        str_len=str_len,
        optimal_value=None,
        termination_error_value=0,
        no_increase_num=no_increase_num,
        minimization=False,
        show_progress_each=1,
        keep_history=True,
    )

    optimizer.fit()

    assert optimizer.get_remains_calls() == pop_size * (iters - no_increase_num - 1)
    assert optimizer._sign == 1

    # start with the optimal_value is equal 1
    optimizer = SHAGA(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        str_len=str_len,
        optimal_value=1,
        termination_error_value=0,
        no_increase_num=None,
        minimization=False,
        show_progress_each=1,
        keep_history=True,
    )

    optimizer.fit()
    assert optimizer.get_remains_calls() == pop_size * (iters - 1)



import numpy as np

import pytest

from sklearn.utils.estimator_checks import check_estimator

import torch

from ..optimizers import SelfCGA
from ..optimizers import SelfCGP
from ..optimizers import GeneticProgramming
from ..optimizers import DifferentialEvolution
from ..regressors import GeneticProgrammingNeuralNetRegressor
from ..regressors import GeneticProgrammingRegressor
from ..regressors import MLPEARegressor


def test_GeneticProgrammingNeuralNetRegressor():

    def problem(x):
        return np.sin(x[:, 0])

    function = problem
    left_border = -4.5
    right_border = 4.5
    sample_size = 300
    n_dimension = 1

    X = np.array(
        [np.linspace(left_border, right_border, sample_size) for _ in range(n_dimension)]
    ).T
    y = function(X)

    model = GeneticProgrammingNeuralNetRegressor(
        n_iter=3,
        pop_size=10,
        optimizer=GeneticProgramming,
        optimizer_args={
            "tour_size": 15,
            "show_progress_each": 1,
            "mutation_rate": 0.07,
            "parents_num": 1,
            "elitism": False,
        },
        weights_optimizer=DifferentialEvolution,
        weights_optimizer_args={"iters": 25, "pop_size": 25, "CR": 0.9},
        max_hidden_block_size=2,
        offset=False,
        test_sample_ratio=0.3,
        net_size_penalty=0.01,
    )

    model.fit(X, y)
    model.predict(X)

    net = model.get_net()
    stats = model.get_stats()
    tree = model.get_tree()

    model = GeneticProgrammingNeuralNetRegressor(
        n_iter=3,
        pop_size=10,
        optimizer=GeneticProgramming,
        optimizer_args={
            "iters": 15,
            "tour_size": 15,
            "show_progress_each": 1,
            "mutation_rate": 0.07,
            "parents_num": 1,
            "elitism": False,
        },
        weights_optimizer=DifferentialEvolution,
        weights_optimizer_args={"iters": 25, "pop_size": 25, "CR": 0.9},
    )

    with pytest.raises(AssertionError):
        model.fit(X, y)

    model = GeneticProgrammingNeuralNetRegressor(
        n_iter=3,
        pop_size=10,
        optimizer=GeneticProgramming,
        optimizer_args={
            "tour_size": 15,
            "show_progress_each": 1,
            "mutation_rate": 0.07,
            "parents_num": 1,
            "elitism": False,
        },
        weights_optimizer=DifferentialEvolution,
        weights_optimizer_args={"iters": 25, "pop_size": 25, "CR": 0.9, "fitness_function": sum},
    )

    with pytest.raises(AssertionError):
        model.fit(X, y)

    model = GeneticProgrammingNeuralNetRegressor(
        n_iter=1,
        pop_size=10,
        input_block_size=10,
        optimizer_args={"show_progress_each": 1},
        weights_optimizer_args={
            "iters": 100,
            "pop_size": 100,
            "no_increase_num": 20,
        },
    )

    check_estimator(model)


def test_SymbolicRegressionGP():

    def problem(x):
        return np.sin(x[:, 0])

    function = problem
    left_border = -4.5
    right_border = 4.5
    sample_size = 300
    n_dimension = 1

    X = np.array(
        [np.linspace(left_border, right_border, sample_size) for _ in range(n_dimension)]
    ).T
    y = function(X)

    model = GeneticProgrammingRegressor(
        n_iter=15,
        pop_size=15,
        functional_set_names=("cos", "sin"),
        optimizer=SelfCGP,
        optimizer_args={"keep_history": True, "show_progress_each": 10, "elitism": True, "K": 2.5},
    )

    model.fit(X, y)
    model.predict(X)

    model = GeneticProgrammingRegressor(
        n_iter=15,
        pop_size=15,
        functional_set_names=("cos", "sin"),
        optimizer=SelfCGP,
        optimizer_args={
            "pop_size": 100,
            "keep_history": True,
            "show_progress_each": 10,
            "elitism": True,
            "K": 2.5,
        },
    )

    with pytest.raises(AssertionError):
        model.fit(X, y)

    model = GeneticProgrammingRegressor(
        n_iter=15,
        pop_size=15,
        functional_set_names=("cos2", "sin"),
        optimizer=SelfCGP,
        optimizer_args={
            "keep_history": True,
            "show_progress_each": 10,
            "elitism": True,
            "K": 2.5,
        },
    )

    with pytest.raises(ValueError):
        model.fit(X, y)

    model = GeneticProgrammingRegressor(
        n_iter=100, pop_size=100, optimizer_args={"no_increase_num": 30}
    )

    check_estimator(model)


def test_MLPEARegressor():
    def problem(x):
        return np.sin(x[:, 0])

    iters = 10
    pop_size = 50

    function = problem
    left_border = -4.5
    right_border = 4.5
    sample_size = 300
    n_dimension = 1

    X = np.array(
        [np.linspace(left_border, right_border, sample_size) for _ in range(n_dimension)]
    ).T
    y = function(X)
    X = torch.as_tensor(X, dtype=torch.float32)
    y = torch.as_tensor(y, dtype=torch.float32)

    model = MLPEARegressor(
        n_iter=15,
        pop_size=15,
        hidden_layers=(5, 3, 25),
        weights_optimizer=SelfCGA,
        weights_optimizer_args={"K": 0.9},
        activation="relu",
        offset=False,
    )

    model.fit(X, y)
    model.predict(X)

    net = model.get_net()
    stat = model.get_stats()

    model = MLPEARegressor(
        n_iter=15,
        pop_size=15,
        hidden_layers=(5, 3, 25),
        weights_optimizer=SelfCGA,
        weights_optimizer_args={"iters": 100, "CR": 0.9},
        activation="relu",
        offset=False,
    )

    with pytest.raises(AssertionError):
        model.fit(X, y)

    model = MLPEARegressor(hidden_layers=(0,), weights_optimizer_args={"no_increase_num": 30})
    check_estimator(model)



from __future__ import annotations

from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np
from numpy.typing import NDArray

from ._geneticprogramming import GeneticProgramming
from ._selfcga import SelfCGA
from ..base import UniversalSet


class SelfCGP(GeneticProgramming, SelfCGA):
    """Semenkin, Eugene & Semenkina, Maria. (2012). Self-configuring genetic programming
    algorithm with modified uniform crossover. 1-6. http://dx.doi.org/10.1109/CEC.2012.6256587"""

    def __init__(
        self,
        fitness_function: Callable[[NDArray[Any]], NDArray[np.float64]],
        uniset: UniversalSet,
        iters: int,
        pop_size: int,
        tour_size: int = 2,
        mutation_rate: float = 0.05,
        parents_num: int = 2,
        elitism: bool = True,
        selections: Tuple[str, ...] = (
            "proportional",
            "rank",
            "tournament_3",
            "tournament_5",
            "tournament_7",
        ),
        crossovers: Tuple[str, ...] = ("gp_standard", "gp_one_point", "gp_uniform_rank_2"),
        mutations: Tuple[str, ...] = (
            "gp_weak_point",
            "gp_average_point",
            "gp_strong_point",
            "gp_weak_grow",
            "gp_average_grow",
            "gp_strong_grow",
        ),
        max_level: int = 16,
        init_level: int = 4,
        init_population: Optional[NDArray] = None,
        K: float = 2,
        selection_threshold_proba: float = 0.05,
        crossover_threshold_proba: float = 0.05,
        mutation_threshold_proba: float = 0.05,
        genotype_to_phenotype: Optional[Callable[[NDArray], NDArray[Any]]] = None,
        optimal_value: Optional[float] = None,
        termination_error_value: float = 0.0,
        no_increase_num: Optional[int] = None,
        minimization: bool = False,
        show_progress_each: Optional[int] = None,
        keep_history: bool = False,
        n_jobs: int = 1,
        fitness_function_args: Optional[Dict] = None,
        genotype_to_phenotype_args: Optional[Dict] = None,
        random_state: Optional[Union[int, np.random.RandomState]] = None,
        on_generation: Optional[Callable] = None,
    ):
        SelfCGA.__init__(
            self,
            fitness_function=fitness_function,
            iters=iters,
            pop_size=pop_size,
            str_len=1,
            tour_size=tour_size,
            mutation_rate=mutation_rate,
            parents_num=parents_num,
            elitism=elitism,
            selections=selections,
            crossovers=crossovers,
            mutations=mutations,
            init_population=init_population,
            K=K,
            selection_threshold_proba=selection_threshold_proba,
            crossover_threshold_proba=crossover_threshold_proba,
            mutation_threshold_proba=mutation_threshold_proba,
            genotype_to_phenotype=genotype_to_phenotype,
            optimal_value=optimal_value,
            termination_error_value=termination_error_value,
            no_increase_num=no_increase_num,
            minimization=minimization,
            show_progress_each=show_progress_each,
            keep_history=keep_history,
            n_jobs=n_jobs,
            fitness_function_args=fitness_function_args,
            genotype_to_phenotype_args=genotype_to_phenotype_args,
            random_state=random_state,
            on_generation=on_generation,
        )

        GeneticProgramming.__init__(
            self,
            fitness_function=fitness_function,
            uniset=uniset,
            iters=iters,
            pop_size=pop_size,
            tour_size=tour_size,
            mutation_rate=mutation_rate,
            parents_num=parents_num,
            elitism=elitism,
            max_level=max_level,
            init_level=init_level,
            init_population=init_population,
            genotype_to_phenotype=genotype_to_phenotype,
            optimal_value=optimal_value,
            termination_error_value=termination_error_value,
            no_increase_num=no_increase_num,
            minimization=minimization,
            show_progress_each=show_progress_each,
            keep_history=keep_history,
            n_jobs=n_jobs,
            fitness_function_args=fitness_function_args,
            genotype_to_phenotype_args=genotype_to_phenotype_args,
            random_state=random_state,
            on_generation=on_generation,
        )



from numba import float64
from numba import int64
from numba import njit

import numpy as np
from numpy.typing import NDArray

import torch


@njit(float64(float64[:], float64[:]))
def root_mean_square_error(
    y_true: NDArray[np.float64], y_predict: NDArray[np.float64]
) -> np.float64:
    error = y_true - y_predict
    mean_squared_error = np.mean(error**2)
    rmse = np.sqrt(mean_squared_error)

    return rmse


@njit(float64[:](float64[:], float64[:, :]))
def root_mean_square_error2d(
    y_true: NDArray[np.float64], y_predict2d: NDArray[np.float64]
) -> NDArray[np.float64]:
    size = len(y_predict2d)
    rmse_values = np.empty(size, dtype=np.float64)

    for i in range(size):
        rmse_values[i] = root_mean_square_error(y_true, y_predict2d[i])

    return rmse_values


@njit(float64(float64[:], float64[:]))
def coefficient_determination(
    y_true: NDArray[np.float64], y_predict: NDArray[np.float64]
) -> np.float64:
    mean_y_true = np.mean(y_true)
    total_sum = np.sum((y_true - mean_y_true) ** 2)

    if total_sum == 0:
        total_sum = 1e-10

    error = y_true - y_predict
    residual_sum = np.sum((error) ** 2)
    r2 = 1 - residual_sum / total_sum

    return r2


@njit(float64[:](float64[:], float64[:, :]))
def coefficient_determination2d(
    y_true: NDArray[np.float64], y_predict2d: NDArray[np.float64]
) -> NDArray[np.float64]:
    size = len(y_predict2d)
    r2_values = np.empty(size, dtype=np.float64)

    for i in range(size):
        r2_values[i] = coefficient_determination(y_true, y_predict2d[i])

    return r2_values


@njit(float64(float64[:, :], float64[:, :]))
def categorical_crossentropy(
    target: NDArray[np.float64], output: NDArray[np.float64]
) -> np.float64:
    target_clipped = np.clip(target, 1e-7, 1 - 1e-7)
    output_clipped = np.clip(output, 1e-7, 1 - 1e-7)

    log_prob = np.log(output_clipped)
    neg_log_prob = -target_clipped * log_prob
    cross_entropy = np.mean(np.sum(neg_log_prob, axis=1))

    return cross_entropy


@njit(float64[:](float64[:, :], float64[:, :, :]))
def categorical_crossentropy3d(
    target: NDArray[np.float64], output3d: NDArray[np.float64]
) -> NDArray[np.float64]:
    size = len(output3d)
    cross_entropy_values = np.empty(size, dtype=np.float64)

    for i in range(size):
        cross_entropy_values[i] = categorical_crossentropy(target, output3d[i])

    return cross_entropy_values


@njit(float64(int64[:], int64[:]))
def accuracy_score(y_true: NDArray[np.int64], y_predict: NDArray[np.int64]) -> np.float64:
    comparison = y_true == y_predict
    comparison_int = comparison.astype(np.int64)
    accuracy = np.mean(comparison_int)

    return accuracy


@njit(float64[:](int64[:], int64[:, :]))
def accuracy_score2d(
    y_true: NDArray[np.int64], y_predict2d: NDArray[np.int64]
) -> NDArray[np.float64]:
    size = len(y_predict2d)
    accuracy_values = np.empty(size, dtype=np.float64)

    for i in range(size):
        accuracy_values[i] = accuracy_score(y_true, y_predict2d[i])

    return accuracy_values


@njit(int64[:, :](int64[:], int64[:]))
def confusion_matrix(y_true: NDArray[np.int64], y_predict: NDArray[np.int64]) -> NDArray[np.int64]:
    classes = np.unique(y_true)
    n_classes = len(classes)

    confusion = np.zeros(shape=(n_classes, n_classes), dtype=np.int64)
    for true, pred in zip(y_true, y_predict):
        confusion[true, pred] += 1

    return confusion


@njit(float64(int64[:], int64[:]))
def recall_score(y_true: NDArray[np.int64], y_predict: NDArray[np.int64]) -> np.float64:
    n_classes = len(np.unique(y_true))
    size = len(y_true)

    true_positives = np.zeros(shape=n_classes, dtype=np.int64)
    false_negatives = np.zeros(shape=n_classes, dtype=np.int64)
    class_recalls = np.zeros(shape=n_classes, dtype=np.float64)

    for i in range(size):
        if y_true[i] == y_predict[i]:
            true_positives[y_true[i]] += 1
        else:
            false_negatives[y_true[i]] += 1

    for i in range(n_classes):
        if true_positives[i] != 0:
            class_recalls[i] = true_positives[i] / (false_negatives[i] + true_positives[i])

    average_recall = np.mean(class_recalls)
    return average_recall


@njit(float64[:](int64[:], int64[:, :]))
def recall_score2d(
    y_true: NDArray[np.int64], y_predict2d: NDArray[np.int64]
) -> NDArray[np.float64]:
    size = len(y_predict2d)
    average_recall_values = np.empty(size, dtype=np.float64)

    for i in range(size):
        average_recall_values[i] = recall_score(y_true, y_predict2d[i])

    return average_recall_values


@njit(float64(int64[:], int64[:]))
def precision_score(y_true: NDArray[np.int64], y_predict: NDArray[np.int64]) -> np.float64:
    classes = np.unique(y_true)
    n_classes = len(classes)
    size = len(y_true)

    true_positives = np.zeros(shape=n_classes, dtype=np.int64)
    false_negatives = np.zeros(shape=n_classes, dtype=np.int64)
    class_precision = np.zeros(shape=n_classes, dtype=np.float64)

    for i in range(size):
        if y_true[i] == y_predict[i]:
            true_positives[y_true[i]] += 1
        else:
            false_negatives[y_predict[i]] += 1

    for i in range(n_classes):
        if true_positives[i] != 0:
            class_precision[i] = true_positives[i] / (false_negatives[i] + true_positives[i])

    average_precision = np.mean(class_precision)
    return average_precision


@njit(float64[:](int64[:], int64[:, :]))
def precision_score2d(
    y_true: NDArray[np.int64], y_predict2d: NDArray[np.int64]
) -> NDArray[np.float64]:
    size = len(y_predict2d)
    average_precision_values = np.empty(size, dtype=np.float64)

    for i in range(size):
        average_precision_values[i] = precision_score(y_true, y_predict2d[i])

    return average_precision_values


@njit(float64(int64[:], int64[:]))
def f1_score(y_true: NDArray[np.int64], y_predict: NDArray[np.int64]) -> np.float64:
    n_classes = len(np.unique(y_true))
    size = len(y_true)

    true_positives = np.zeros(shape=n_classes, dtype=np.int64)
    false_negatives = np.zeros(shape=n_classes, dtype=np.int64)
    down_precision = np.zeros(shape=n_classes, dtype=np.int64)
    f1 = np.zeros(shape=n_classes, dtype=np.float64)

    for i in range(size):
        if y_true[i] == y_predict[i]:
            true_positives[y_true[i]] += 1
        else:
            false_negatives[y_true[i]] += 1
            down_precision[y_predict[i]] += 1

    for i in range(n_classes):
        if true_positives[i] != 0:
            precision = true_positives[i] / (down_precision[i] + true_positives[i])
            recall = true_positives[i] / (false_negatives[i] + true_positives[i])
            f1[i] = 2 * (precision * recall) / (precision + recall)

    average_f1 = np.mean(f1)
    return average_f1


@njit(float64[:](int64[:], int64[:, :]))
def f1_score2d(y_true: NDArray[np.int64], y_predict2d: NDArray[np.int64]) -> NDArray[np.float64]:
    size = len(y_predict2d)
    average_f1_values = np.empty(size, dtype=np.float64)

    for i in range(size):
        average_f1_values[i] = f1_score(y_true, y_predict2d[i])

    return average_f1_values


def _ce_from_probs_torch(
    probs: torch.Tensor, targets: torch.Tensor, eps: float = 1e-12
) -> torch.Tensor:

    if probs.ndim == 3 and probs.size(-1) == 1:
        probs = probs.squeeze(-1)
    if targets.ndim == 3 and targets.size(-1) == 1:
        targets = targets.squeeze(-1)
    if probs.ndim == 3:
        probs = probs.reshape(-1, probs.size(-1))
        targets = targets.reshape(-1, targets.size(-1))

    probs = probs.clamp_min(eps)
    loss_vec = -(targets * probs.log()).sum(dim=-1)
    return loss_vec.mean()


def _rmse_torch(pred: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
    if pred.ndim == 3 and pred.size(-1) == 1:
        pred = pred.squeeze(-1)
    if target.ndim == 3 and target.size(-1) == 1:
        target = target.squeeze(-1)
    if pred.ndim == 2 and pred.size(-1) == 1:
        pred = pred.squeeze(-1)
    if target.ndim == 2 and target.size(-1) == 1:
        target = target.squeeze(-1)
    return torch.sqrt(torch.mean((pred - target) ** 2))



import numpy as np

from ._optproblems import ShiftedSphere  # 1 Unimodal (5)
from ._optproblems import ShiftedSchwefe1_2  # 2
from ._optproblems import ShiftedRotatedHighConditionedElliptic  # 3
from ._optproblems import ShiftedSchwefe1_2WithNoise  # 4
from ._optproblems import Schwefel2_6  # 5

from ._optproblems import ShiftedRosenbrock  # 6 Multimodal (20) Basic Functions (7)
from ._optproblems import ShiftedRotatedGriewank  # 7
from ._optproblems import ShiftedRotatedAckley  # 8
from ._optproblems import ShiftedRastrigin  # 9
from ._optproblems import ShiftedRotatedRastrigin  # 10
from ._optproblems import ShiftedRotatedWeierstrass  # 11
from ._optproblems import Schwefel2_13  # 12

from ._optproblems import ShiftedExpandedGriewankRosenbrock  # 13 Expanded Functions (2)
from ._optproblems import ShiftedRotatedExpandedScaffes_F6  # 14

from ._optproblems import HybridCompositionFunction1  # 15 Hybrid Composition Functions (11)
from ._optproblems import RotatedVersionHybridCompositionFunction1  # 16
from ._optproblems import RotatedVersionHybridCompositionFunction1Noise  # 17
from ._optproblems import RotatedHybridCompositionFunction  # 18
from ._optproblems import RotatedHybridCompositionFunctionNarrowBasin  # 19
from ._optproblems import RotatedHybridCompositionFunctionOptimalBounds  # 20
from ._optproblems import HybridCompositionFunction3  # 21
from ._optproblems import HybridCompositionFunction3H  # 22
from ._optproblems import NonContinuousHybridCompositionFunction3  # 23
from ._optproblems import HybridCompositionFunction4  # 24
from ._optproblems import HybridCompositionFunction4withoutbounds  # 25


"""Suganthan, Ponnuthurai & Hansen, Nikolaus & Liang, Jing & Deb, Kalyan & Chen, Ying-ping & Auger,
Anne & Tiwari, Santosh. (2005). Problem Definitions and Evaluation Criteria for the CEC 2005 Special
Session on Real-Parameter Optimization. Natural Computing. 341-357"""

problems_dict = {
    "F1": {
        "function": ShiftedSphere,
        "bounds": (-100, 100),
        "fix_accuracy": 1e-6,
        "optimum": -450,
        "optimum_x": ShiftedSphere().x_shift,
        "dimentions": range(2, 101),
    },
    "F2": {
        "function": ShiftedSchwefe1_2,
        "bounds": (-100, 100),
        "fix_accuracy": 1e-6,
        "optimum": -450,
        "optimum_x": ShiftedSchwefe1_2().x_shift,
        "dimentions": range(2, 101),
    },
    "F3": {
        "function": ShiftedRotatedHighConditionedElliptic,
        "bounds": (-100, 100),
        "fix_accuracy": 1e-6,
        "optimum": -450,
        "optimum_x": ShiftedRotatedHighConditionedElliptic().x_shift,
        "dimentions": (2, 10, 30, 50),
    },
    "F4": {
        "function": ShiftedSchwefe1_2WithNoise,
        "bounds": (-100, 100),
        "fix_accuracy": 1e-6,
        "optimum": -450,
        "optimum_x": ShiftedSchwefe1_2WithNoise().x_shift,
        "dimentions": range(2, 101),
    },
    "F5": {
        "function": Schwefel2_6,
        "bounds": (-100, 100),
        "fix_accuracy": 1e-6,
        "optimum": -310,
        "optimum_x": Schwefel2_6().x_shift,
        "dimentions": range(2, 101),
    },
    "F6": {
        "function": ShiftedRosenbrock,
        "bounds": (-100, 100),
        "fix_accuracy": 1e-2,
        "optimum": 390,
        "optimum_x": ShiftedRosenbrock().x_shift,
        "dimentions": range(2, 101),
    },
    "F7": {
        "function": ShiftedRotatedGriewank,
        "bounds": (-1000, 1000),
        "fix_accuracy": 1e-2,
        "optimum": -180,
        "optimum_x": ShiftedRotatedGriewank().x_shift,
        "init_bounds": (0, 600),
        "dimentions": (2, 10, 30, 50),
    },
    "F8": {
        "function": ShiftedRotatedAckley,
        "bounds": (-32, 32),
        "fix_accuracy": 1e-2,
        "optimum": -140,
        "optimum_x": ShiftedRotatedAckley().x_shift,
        "dimentions": (2, 10, 30, 50),
    },
    "F9": {
        "function": ShiftedRastrigin,
        "bounds": (-5, 5),
        "fix_accuracy": 1e-2,
        "optimum": -330,
        "optimum_x": ShiftedRastrigin().x_shift,
        "dimentions": range(2, 51),
    },
    "F10": {
        "function": ShiftedRotatedRastrigin,
        "bounds": (-5, 5),
        "fix_accuracy": 1e-2,
        "optimum": -330,
        "optimum_x": ShiftedRotatedRastrigin().x_shift,
        "dimentions": (2, 10, 30, 50),
    },
    "F11": {
        "function": ShiftedRotatedWeierstrass,
        "bounds": (-0.5, 0.5),
        "fix_accuracy": 1e-2,
        "optimum": 90,
        "optimum_x": ShiftedRotatedWeierstrass().x_shift,
        "dimentions": (2, 10, 30, 50),
    },
    "F12": {
        "function": Schwefel2_13,
        "bounds": (-np.pi, np.pi),
        "fix_accuracy": 1e-2,
        "optimum": -460,
        "optimum_x": Schwefel2_13().x_shift,
        "dimentions": range(2, 101),
    },
    "F13": {
        "function": ShiftedExpandedGriewankRosenbrock,
        "bounds": (-3, 1),
        "fix_accuracy": 1e-2,
        "optimum": -130,
        "optimum_x": ShiftedExpandedGriewankRosenbrock().x_shift,
        "dimentions": range(2, 101),
    },
    "F14": {
        "function": ShiftedRotatedExpandedScaffes_F6,
        "bounds": (-100, 100),
        "fix_accuracy": 1e-2,
        "optimum": -300,
        "optimum_x": ShiftedRotatedExpandedScaffes_F6().x_shift,
        "dimentions": (2, 10, 30, 50),
    },
    "F15": {
        "function": HybridCompositionFunction1,
        "bounds": (-5, 5),
        "fix_accuracy": 1e-2,
        "optimum": 120,
        "optimum_x": HybridCompositionFunction1().x_shift[0],
        "dimentions": (2, 10, 30, 50),
    },
    "F16": {
        "function": RotatedVersionHybridCompositionFunction1,
        "bounds": (-5, 5),
        "fix_accuracy": 1e-2,
        "optimum": 120,
        "optimum_x": RotatedVersionHybridCompositionFunction1().x_shift[0],
        "dimentions": (2, 10, 30, 50),
    },
    "F17": {
        "function": RotatedVersionHybridCompositionFunction1Noise,
        "bounds": (-5, 5),
        "fix_accuracy": 1e-1,
        "optimum": 120,
        "optimum_x": RotatedVersionHybridCompositionFunction1Noise().x_shift[0],
        "dimentions": (2, 10, 30, 50),
    },
    "F18": {
        "function": RotatedHybridCompositionFunction,
        "bounds": (-5, 5),
        "fix_accuracy": 1e-1,
        "optimum": 10,
        "optimum_x": RotatedHybridCompositionFunction().x_shift[0],
        "dimentions": (2, 10, 30, 50),
    },
    "F19": {
        "function": RotatedHybridCompositionFunctionNarrowBasin,
        "bounds": (-5, 5),
        "fix_accuracy": 1e-1,
        "optimum": 10,
        "optimum_x": RotatedHybridCompositionFunctionNarrowBasin().x_shift[0],
        "dimentions": (2, 10, 30, 50),
    },
    "F20": {
        "function": RotatedHybridCompositionFunctionOptimalBounds,
        "bounds": (-5, 5),
        "fix_accuracy": 1e-1,
        "optimum": 10,
        "optimum_x": RotatedHybridCompositionFunctionOptimalBounds().x_shift[0],
        "dimentions": (2, 10, 30, 50),
    },
    "F21": {
        "function": HybridCompositionFunction3,
        "bounds": (-5, 5),
        "fix_accuracy": 1e-1,
        "optimum": 360,
        "optimum_x": HybridCompositionFunction3().x_shift[0],
        "dimentions": (2, 10, 30, 50),
    },
    "F22": {
        "function": HybridCompositionFunction3H,
        "bounds": (-5, 5),
        "fix_accuracy": 1e-1,
        "optimum": 360,
        "optimum_x": HybridCompositionFunction3H().x_shift[0],
        "dimentions": (2, 10, 30, 50),
    },
    "F23": {
        "function": NonContinuousHybridCompositionFunction3,
        "bounds": (-5, 5),
        "fix_accuracy": 1e-1,
        "optimum": 360,
        "optimum_x": NonContinuousHybridCompositionFunction3().x_shift[0],
        "dimentions": (2, 10, 30, 50),
    },
    "F24": {
        "function": HybridCompositionFunction4,
        "bounds": (-5, 5),
        "fix_accuracy": 1e-1,
        "optimum": 260,
        "optimum_x": HybridCompositionFunction4().x_shift[0],
        "dimentions": (2, 10, 30, 50),
    },
    "F25": {
        "function": HybridCompositionFunction4withoutbounds,
        "bounds": (-10, 10),
        "fix_accuracy": 1e-1,
        "optimum": 260,
        "init_bounds": (2, 5),
        "optimum_x": HybridCompositionFunction4withoutbounds().x_shift[0],
        "dimentions": (2, 10, 30, 50),
    },
}



from __future__ import annotations

from typing import Any
from typing import Optional
from typing import Type
from typing import Union

import numpy as np
from numpy.typing import NDArray
from numpy.typing import ArrayLike

import torch

from sklearn.base import ClassifierMixin
from sklearn.utils.validation import check_is_fitted
from sklearn.utils.validation import validate_data

from ..base._gpnn import BaseGPNN
from ..base._mlp import weights_type_optimizer_alias
from ..optimizers import GeneticProgramming
from ..optimizers import SHADE
from ..optimizers import SelfCGP


class GeneticProgrammingNeuralNetClassifier(ClassifierMixin, BaseGPNN):
    def __init__(
        self,
        *,
        n_iter: int = 15,
        pop_size: int = 50,
        input_block_size: int = 1,
        max_hidden_block_size: int = 9,
        offset: bool = True,
        test_sample_ratio: float = 0.5,
        optimizer: Union[Type[SelfCGP], Type[GeneticProgramming]] = SelfCGP,
        optimizer_args: Optional[dict[str, Any]] = None,
        weights_optimizer: weights_type_optimizer_alias = SHADE,
        weights_optimizer_args: Optional[dict[str, Any]] = None,
        net_size_penalty: float = 0.0,
        random_state: Optional[Union[int, np.random.RandomState]] = None,
        device: str = "cpu",
        use_fitness_cache: bool = False,
        fitness_cache_size: int = 1000,
    ):
        super().__init__(
            n_iter=n_iter,
            pop_size=pop_size,
            input_block_size=input_block_size,
            max_hidden_block_size=max_hidden_block_size,
            offset=offset,
            test_sample_ratio=test_sample_ratio,
            optimizer=optimizer,
            optimizer_args=optimizer_args,
            weights_optimizer=weights_optimizer,
            weights_optimizer_args=weights_optimizer_args,
            net_size_penalty=net_size_penalty,
            random_state=random_state,
            device=device,
            use_fitness_cache=use_fitness_cache,
            fitness_cache_size=fitness_cache_size,
        )

    def predict_proba(self, X: ArrayLike) -> NDArray[np.float64]:
        check_is_fitted(self)

        if hasattr(self, "_validate_data"):
            X = self._validate_data(X, reset=False)
        else:
            X = validate_data(self, X, reset=False)

        if self.offset:
            X = np.hstack([X, np.ones((X.shape[0], 1))])

        device = torch.device(self.device)
        X_t = torch.as_tensor(X, dtype=torch.float64, device=device)

        with torch.no_grad():
            proba_t = self.net_.forward(X_t)
        proba = proba_t.detach().cpu().to(torch.float64).numpy()
        return np.ascontiguousarray(proba, dtype=np.float64)

    def predict(self, X: ArrayLike):
        proba = self.predict_proba(X)
        indeces = np.argmax(proba, axis=1)

        return self._label_encoder.inverse_transform(indeces)



from __future__ import annotations

from collections import defaultdict
from itertools import product
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple

import hashlib

import numpy as np
from numpy.typing import NDArray

from ..utils import forward2d
from ..utils.transformations import minmax_scale
from ..utils.random import random_sample
from ..utils.random import uniform

import torch
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

SIGMA, RELU, GAUSS, TANH, LN, SOFTMAX = 0, 1, 2, 3, 4, 5


def _act(code: int, x: torch.Tensor) -> torch.Tensor:
    if code == SIGMA:
        return torch.sigmoid(x)
    if code == RELU:
        return F.relu(x)
    if code == GAUSS:
        return torch.exp(-x * x)
    if code == TANH:
        return torch.tanh(x)
    if code == LN:
        return x
    if code == SOFTMAX:
        return F.softmax(x, dim=1)  # softmax по оси узлов
    raise ValueError(f"Unknown activation code {code}")


ACTIVATION_NAME = {0: "sg", 1: "rl", 2: "gs", 3: "th", 4: "ln", 5: "sm"}
ACTIV_NAME_INV = {"sigma": 0, "relu": 1, "gauss": 2, "tanh": 3, "ln": 4, "softmax": 5}


class _Block(nn.Module):
    """Контейнер для индексов одного топологического шага (только буферы, веса не дублируем)."""

    def __init__(
        self,
        fr: np.ndarray,
        to: np.ndarray,
        widx: np.ndarray,
        act_codes: np.ndarray,
        act_nodes_list: list[np.ndarray],
    ):
        super().__init__()
        self.register_buffer("from_idx", torch.as_tensor(fr, dtype=torch.long))
        self.register_buffer("to_idx", torch.as_tensor(to, dtype=torch.long))
        self.register_buffer("weight_idx", torch.as_tensor(widx, dtype=torch.long)) 
        self.register_buffer("act_codes", torch.as_tensor(act_codes, dtype=torch.long))
        lens = [len(a) for a in act_nodes_list]
        maxlen = max(lens) if lens else 0
        pad = torch.full((len(act_nodes_list), maxlen), -1, dtype=torch.long)
        for i, arr in enumerate(act_nodes_list):
            if len(arr):
                pad[i, : len(arr)] = torch.as_tensor(arr, dtype=torch.long)
        self.register_buffer("act_nodes", pad)
        self.register_buffer("act_nodes_len", torch.as_tensor(lens, dtype=torch.long))


class Net:
    def __init__(
        self,
        inputs: Optional[Set] = None,
        hidden_layers: Optional[List] = None,
        outputs: Optional[Set] = None,
        connects: Optional[NDArray[np.int64]] = None,
        weights: Optional[torch.Tensor] = None,
        activs: Optional[Dict[int, int]] = None,
    ):
        self._inputs = inputs or set()
        self._hidden_layers = hidden_layers or []
        self._outputs = outputs or set()
        self._connects = self._set_connects(values=connects)
        self._weights = self._set_weights(values=weights)
        self._activs = activs or {}
        self._offset: bool = False

        self._numpy_inputs: Optional[NDArray[np.int64]] = None
        self._numpy_outputs: NDArray[np.int64]
        self._n_hiddens: np.int64

        self.blocks: list[_Block] | None = None
        self.inputs: torch.Tensor | None = None
        self.outputs: torch.Tensor | None = None
        self.n_nodes: int | None = None

        self._activation_name = {0: "sg", 1: "rl", 2: "gs", 3: "th", 4: "ln", 5: "sm"}
        self._activ_name_inv = {"sigma": 0, "relu": 1, "gauss": 2, "tanh": 3, "ln": 4, "softmax": 5}

    def to(self, device: str | torch.device, dtype: Optional[torch.dtype] = None) -> "Net":
        dev = torch.device(device)
        self._weights = self._weights.to(device=dev if dtype is None else dev, dtype=dtype)

        if self.inputs is not None:
            self.inputs = self.inputs.to(dev)
        if self.outputs is not None:
            self.outputs = self.outputs.to(dev)
        if self.blocks is not None:
            for b in self.blocks:
                b.to(dev)
        return self

    def cuda(self) -> Net:
        return self.to("cuda")

    def cpu(self) -> Net:
        return self.to("cpu")

    def __len__(self) -> int:
        return int(self._weights.numel())

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Net):
            return NotImplemented

        if self._inputs != other._inputs:
            return False

        if len(self._hidden_layers) != len(other._hidden_layers):
            return False
        if any(h1 != h2 for h1, h2 in zip(self._hidden_layers, other._hidden_layers)):
            return False

        if self._outputs != other._outputs:
            return False

        if set(map(tuple, self._connects)) != set(map(tuple, other._connects)):
            return False

        if self._activs != other._activs:
            return False

        return True

    def _set_connects(self, values: Optional[NDArray[np.int64]]) -> NDArray[np.int64]:
        if values is None:
            to_return = np.empty((0, 2), dtype=np.int64)
        else:
            to_return = values
        return to_return

    def _set_weights(self, values: Optional[torch.Tensor]) -> torch.Tensor:
        if values is None:
            return torch.empty(0, dtype=torch.float32, device="cpu")
        return values.to(dtype=torch.float32)

    def copy(self) -> "Net":
        hidden_layers = [layer.copy() for layer in self._hidden_layers]
        copy_net = Net(
            inputs=self._inputs.copy(),
            hidden_layers=hidden_layers,
            outputs=self._outputs.copy(),
            connects=self._connects.copy(),
            weights=self._weights.detach().clone().to(self._weights.device),
            activs=self._activs.copy(),
        )
        copy_net._offset = self._offset
        return copy_net

    def _assemble_hiddens(self) -> Set[int]:
        if len(self._hidden_layers) > 0:
            return set.union(*self._hidden_layers)
        else:
            return set()

    def signature(self) -> str:
        inputs = tuple(sorted(self._inputs))
        outputs = tuple(sorted(self._outputs))
        
        hidden_layers = tuple(tuple(sorted(layer)) for layer in self._hidden_layers)
        
        connects = tuple(sorted(map(tuple, self._connects.tolist())))
        
        activs = tuple(sorted(self._activs.items()))
        
        offset = self._offset
        
        sig_tuple = (inputs, hidden_layers, outputs, connects, activs, offset)
        
        sig_str = repr(sig_tuple).encode()
        return hashlib.sha1(sig_str).hexdigest()

    def _get_connect(
        self, left: Set[int], right: Set[int]
    ) -> Tuple[NDArray[np.int64], NDArray[np.float64]]:
        if len(left) and len(right):
            connects = np.array(list(product(left, right)), dtype=np.int64)
            weights = uniform(-2, 2, len(connects))
            return (connects, weights)
        else:
            return (np.zeros((0, 2), dtype=np.int64), np.zeros((0), dtype=np.float64))

    def __add__(self, other: Net) -> Net:
        len_i_1, len_i_2 = len(self._inputs), len(other._inputs)
        len_h_1, len_h_2 = len(self._hidden_layers), len(other._hidden_layers)

        if (len_i_1 > 0 and len_i_2 == 0) and (len_h_1 == 0 and len_h_2 > 0):
            return self > other
        elif (len_i_1 == 0 and len_i_2 > 0) and (len_h_1 > 0 and len_h_2 == 0):
            return other > self

        map_res = map(
            lambda layers: layers[0].union(layers[1]),
            zip(self._hidden_layers, other._hidden_layers),
        )

        if len_h_1 < len_h_2:
            excess = other._hidden_layers[len_h_1:]
        elif len_h_1 > len_h_2:
            excess = self._hidden_layers[len_h_2:]
        else:
            excess = []

        hidden = list(map_res) + excess
        return Net(
            inputs=self._inputs.union(other._inputs),
            hidden_layers=hidden,
            outputs=self._outputs.union(other._outputs),
            connects=np.vstack([self._connects, other._connects]),
            weights=torch.cat([self._weights, other._weights], dim=0),
            activs={**self._activs, **other._activs},
        )

    def __gt__(self, other: Net) -> Net:
        len_i_1, len_i_2 = len(self._inputs), len(other._inputs)
        len_h_1, len_h_2 = len(self._hidden_layers), len(other._hidden_layers)

        if (len_i_1 > 0 and len_h_1 == 0) and (len_i_2 > 0 and len_h_2 == 0):
            return self + other
        elif (len_i_1 == 0 and len_h_1 > 0) and (len_i_2 > 0 and len_h_2 == 0):
            return other > self

        inputs_hidden = self._inputs.union(self._assemble_hiddens())
        from_ = inputs_hidden.difference(self._connects[:, 0])

        cond = other._connects[:, 0][:, np.newaxis] == np.array(list(other._inputs))
        cond = np.any(cond, axis=1)

        connects_no_i = other._connects[:, 1][~cond]
        hidden_outputs = other._assemble_hiddens().union(other._outputs)
        to_ = hidden_outputs.difference(connects_no_i)

        connects, weights_np = self._get_connect(from_, to_)
        weights_t = torch.as_tensor(weights_np, dtype=torch.float32, device=self._weights.device)
        return Net(
            inputs=self._inputs.union(other._inputs),
            hidden_layers=self._hidden_layers + other._hidden_layers,
            outputs=self._outputs.union(other._outputs),
            connects=np.vstack([self._connects, other._connects, connects]),
            weights=torch.cat([self._weights, other._weights, weights_t], dim=0),
            activs={**self._activs, **other._activs},
        )

    def _fix(self, inputs: Set[int]) -> Net:
        hidden_outputs = self._assemble_hiddens().union(self._outputs)
        to_ = hidden_outputs.difference(self._connects[:, 1])
        if len(to_) > 0:
            if not len(self._inputs):
                self._inputs = inputs
            connects, weights_np = self._get_connect(self._inputs, to_)
            weights_t = torch.as_tensor(
                weights_np, dtype=torch.float32, device=self._weights.device
            )
            self._connects = np.vstack([self._connects, connects])
            self._weights = torch.cat([self._weights, weights_t], dim=0)

        self._connects = np.unique(self._connects, axis=0)
        if self._weights.numel() > self._connects.shape[0]:
            self._weights = self._weights[: self._connects.shape[0]]
        return self

    def _build_order(self) -> list[_Block]:
        hidden = set.union(*self._hidden_layers) if self._hidden_layers else set()
        fr = self._connects[:, 0]
        to = self._connects[:, 1]
        idx = np.arange(len(fr), dtype=np.int64)

        order = np.argsort(to)
        fr_s, to_s, idx_s = fr[order], to[order], idx[order]

        groups_to, cuts = np.unique(to_s, return_index=True)
        groups_from = np.split(fr_s, cuts)[1:]
        group_idx = np.split(idx_s, cuts)[1:]

        pairs: dict[tuple[int, ...], list[int]] = {}
        widx_map: dict[tuple[int, ...], np.ndarray] = {}

        for g_to, g_from, g_i in zip(groups_to, groups_from, group_idx):
            ord2 = np.argsort(g_from)
            f_sorted = tuple(g_from[ord2])
            pairs.setdefault(f_sorted, []).append(int(g_to))
            widx_map.setdefault(f_sorted, []).append(g_i[ord2])

        widx_map = {k: np.asarray(v, dtype=np.int64) for k, v in widx_map.items()}

        calculated = set(self._inputs)
        purpose = set(self._inputs) | hidden | set(self._outputs)

        blocks: list[_Block] = []
        while calculated != purpose:
            progressed = False
            for f_tuple, to_nodes in pairs.items():
                if set(f_tuple).issubset(calculated) and not set(to_nodes).issubset(calculated):
                    progressed = True
                    calculated |= set(to_nodes)

                    nodes_map: dict[int, list[int]] = {}
                    for t in to_nodes:
                        nodes_map.setdefault(self._activs[t], []).append(t)

                    act_codes = np.fromiter(nodes_map.keys(), dtype=np.int64)
                    act_nodes = [np.asarray(v, dtype=np.int64) for v in nodes_map.values()]

                    blocks.append(
                        _Block(
                            fr=np.asarray(f_tuple, dtype=np.int64),
                            to=np.asarray(to_nodes, dtype=np.int64),
                            widx=widx_map[f_tuple],
                            act_codes=act_codes,
                            act_nodes_list=act_nodes,
                        )
                    )
            if not progressed:
                raise RuntimeError("Topological build stalled (cycle or disconnected nodes).")
        return blocks

    def compile_torch(self, device: torch.device | str | None = None) -> "Net":
        dev = torch.device(device) if device is not None else self._weights.device

        all_nodes = set(self._inputs)
        if self._hidden_layers:
            all_nodes |= set.union(*self._hidden_layers)
        all_nodes |= set(self._outputs)
        self.n_nodes = (max(all_nodes) + 1) if all_nodes else 0

        self.inputs = torch.as_tensor(sorted(self._inputs), dtype=torch.long, device=dev)
        self.outputs = torch.as_tensor(sorted(self._outputs), dtype=torch.long, device=dev)

        blocks = self._build_order()
        for b in blocks:
            b.to(dev)
        self.blocks = blocks
        return self

    def ensure_compiled(self) -> None:
        if (
            self.blocks is None
            or self.inputs is None
            or self.outputs is None
            or self.n_nodes is None
        ):
            self.compile_torch()

    def forward(
        self,
        X: torch.Tensor,
        weights: torch.Tensor | None = None,
        keep_weight_dim: bool = False,
        autocast_input: bool = True,
    ) -> torch.Tensor:
        base_w = self._weights if weights is None else weights
        w = base_w.view(1, -1) if base_w.ndim == 1 else base_w
        W = w.shape[0]

        if autocast_input:
            X = X.to(device=w.device, dtype=w.dtype)
        else:
            if X.device != w.device:
                raise RuntimeError(f"X on {X.device}, weights on {w.device}. Move X to {w.device}.")
            if X.dtype != w.dtype:
                raise RuntimeError(f"dtype mismatch: X {X.dtype} vs weights {w.dtype}.")

        B = X.shape[0]

        vals = torch.zeros(W, self.n_nodes, B, device=X.device, dtype=X.dtype)
        if X.shape[1] == self.inputs.numel():
            X_in = X
        else:
            X_in = X.index_select(1, self.inputs)

        vals[:, self.inputs, :] = X_in.T.unsqueeze(0).expand(W, -1, -1)

        for blk in self.blocks:
            wid_flat = blk.weight_idx.reshape(1, -1).expand(W, -1)
            w_blk = w.gather(1, wid_flat).view(W, blk.weight_idx.shape[0], blk.weight_idx.shape[1])

            z = torch.matmul(w_blk, vals[:, blk.from_idx, :])
            vals[:, blk.to_idx, :] = z

            for i in range(blk.act_codes.numel()):
                k = int(blk.act_nodes_len[i])
                if k == 0:
                    continue
                idx = blk.act_nodes[i, :k]
                code = int(blk.act_codes[i])
                vals[:, idx, :] = _act(code, vals[:, idx, :])

        out = vals[:, self.outputs, :].permute(0, 2, 1)
        if W == 1 and not keep_weight_dim:
            out = out.squeeze(0)
        return out

    def get_graph(self) -> Dict:
        input_color_code = (0.11, 0.67, 0.47, 1)
        hidden_color_code = (0.0, 0.74, 0.99, 1)
        output_color_code = (0.94, 0.50, 0.50, 1)

        weights_scale = minmax_scale(self._weights.detach().cpu().numpy())
        nodes = list(self._inputs)

        len_i = len(self._inputs)
        len_h = len(self._assemble_hiddens())
        len_o = len(self._outputs)
        sum_ = len_i + len_h + len_o

        positions = np.zeros((sum_, 2), dtype=float)
        colors = np.zeros((sum_, 4))
        w_colors = np.zeros((len(weights_scale), 4))
        labels = {
            **dict(zip(self._inputs, self._inputs)),
            **{key: ACTIVATION_NAME[value] for key, value in self._activs.items()},
            **dict(zip(self._outputs, range(len_o))),
        }

        w_colors[:, 0] = 1 - weights_scale
        w_colors[:, 2] = weights_scale
        w_colors[:, 3] = 0.8
        positions[:len_i][:, 1] = np.arange(len_i) - (len_i) / 2
        colors[:len_i] = input_color_code

        n = len_i
        for i, layer in enumerate(self._hidden_layers):
            nodes.extend(list(layer))
            positions[n : n + len(layer)][:, 0] = i + 1
            positions[n : n + len(layer)][:, 1] = np.arange(len(layer)) - len(layer) / 2
            colors[n : n + len(layer)] = hidden_color_code
            n += len(layer)

        nodes.extend(list(self._outputs))
        positions[n : n + len_o][:, 0] = len(self._hidden_layers) + 1
        positions[n : n + len_o][:, 1] = np.arange(len_o) - len_o / 2
        colors[n : n + len_o] = output_color_code

        positions_dict = dict(zip(nodes, positions))

        to_return = {
            "nodes": nodes,
            "labels": labels,
            "positions": positions_dict,
            "colors": colors,
            "weights_colors": w_colors,
            "connects": self._connects,
        }

        return to_return

    def plot(self, ax=None) -> None:
        import networkx as nx

        graph = self.get_graph()
        G = nx.Graph()
        G.add_nodes_from(graph["nodes"])
        G.add_edges_from(graph["connects"])

        connects_label = {}
        for i, connects_i in enumerate(graph["connects"]):
            connects_label[tuple(connects_i)] = i

        nx.draw_networkx_nodes(
            G,
            pos=graph["positions"],
            node_color=graph["colors"],
            edgecolors="black",
            linewidths=0.5,
            ax=ax,
        )
        nx.draw_networkx_edges(
            G,
            pos=graph["positions"],
            style="-",
            edge_color=graph["weights_colors"],
            ax=ax,
            alpha=0.5,
        )

        nx.draw_networkx_labels(G, graph["positions"], graph["labels"], font_size=10, ax=ax)


class HiddenBlock:
    def __init__(self, max_size: int) -> None:
        self._activ = random_sample(5, 1, True)[0]
        self._size = random_sample(max_size - 1, 1, True)[0] + 1

    def __str__(self) -> str:
        return "{}{}".format(ACTIVATION_NAME[self._activ], self._size)


class NetEnsemble:
    def __init__(self, nets: NDArray, meta_algorithm: Optional[Net] = None):
        self._nets = nets
        self._meta_algorithm = meta_algorithm
        self._meta_tree = None

    def __len__(self) -> int:
        return len(self._nets)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, NetEnsemble):
            if len(self) != len(other):
                return False
            elif any((net_i != net_j for net_i, net_j in zip(self._nets, other._nets))):
                return False
            else:
                return True
        return NotImplemented

    def get_nets(self: NetEnsemble) -> NDArray:
        return self._nets

    def forward(
        self: NetEnsemble,
        X: NDArray[np.float64],
        weights_list: Optional[List[NDArray[np.float64]]] = None,
    ) -> NDArray[np.float64]:
        if weights_list is not None:
            to_return = [
                net_i.forward(X, weights_i) for net_i, weights_i in zip(self._nets, weights_list)
            ]
        else:
            to_return = [net_i.forward(X) for net_i in self._nets]

        return np.array(to_return, dtype=np.float64)

    def _get_meta_inputs(
        self: NetEnsemble, X: NDArray[np.float64], offset: bool = True
    ) -> NDArray[np.float64]:
        outputs = self.forward(X)[:, 0]
        X_input = np.concatenate(outputs, axis=1)
        if offset:
            X_input = np.hstack([X_input, np.ones((X_input.shape[0], 1))])

        return X_input

    def meta_output(
        self: NetEnsemble,
        X: NDArray[np.float64],
    ) -> NDArray[np.float64]:
        if self._meta_algorithm is not None:
            X_input = self._get_meta_inputs(X, offset=self._meta_algorithm._offset)
            output = self._meta_algorithm.forward(X=X_input)[0]
            return output
        else:
            raise ValueError("text222")

    def copy(self) -> NetEnsemble:
        copy_ = NetEnsemble(np.array([net_i.copy() for net_i in self._nets], dtype=object))
        if self._meta_algorithm is not None:
            copy_._meta_algorithm = self._meta_algorithm.copy()

        if self._meta_tree is not None:
            copy_._meta_tree = self._meta_tree.copy()
        return copy_



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Type
from typing import Union

from collections import OrderedDict

import numpy as np
from numpy.typing import ArrayLike
from numpy.typing import NDArray

from sklearn.base import BaseEstimator
from sklearn.base import ClassifierMixin
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import LabelEncoder
from sklearn.preprocessing import OneHotEncoder
from sklearn.utils.multiclass import check_classification_targets
from sklearn.utils.validation import validate_data

import torch

from ..base import FunctionalNode
from ..base import Net
from ..base import TerminalNode
from ..base import Tree
from ..base import UniversalSet
from ..base._ea import Statistics
from ..base._mlp import check_optimizer_args
from ..base._mlp import fitness_function_weights
from ..base._mlp import train_net_weights
from ..base._mlp import weights_type_optimizer_alias
from ..base._net import ACTIV_NAME_INV
from ..base._tree import init_net_uniset
from ..optimizers import GeneticProgramming
from ..optimizers import SHAGA
from ..optimizers import SelfCGP
from ..utils import array_like_to_numpy_X_y
from ..utils.random import check_random_state
from ..utils._metrics import _ce_from_probs_torch, _rmse_torch


class LimitedSizeDict(OrderedDict):
    def __init__(self, max_size: int, *args, **kwargs):
        self.max_size = max_size
        super().__init__(*args, **kwargs)

    def __setitem__(self, key, value):
        if key in self:
            super().__delitem__(key)
        elif len(self) >= self.max_size:
            self.popitem(last=False)
        super().__setitem__(key, value)

def fitness_function_structure(
    population: NDArray,
    X: torch.Tensor,
    targets: torch.Tensor,
    net_size_penalty: float,
    task_type: str = "regression",
    fitness_cache: Optional[dict[str, dict]] = None,
) -> NDArray[np.float64]:

    losses = np.empty(len(population), dtype=np.float64)
    with torch.no_grad():
        for i, net in enumerate(population):
            sig = net.signature()

            if fitness_cache is not None and sig in fitness_cache:
                net._weights = fitness_cache[sig]["weights"].clone().to(X.device)
                net.compile_torch()

                if fitness_cache[sig]["fitness"] is not None:
                    losses[i] = fitness_cache[sig]["fitness"]
                    continue

            out = net.forward(X)
            if task_type == "classification":
                loss_t = _ce_from_probs_torch(out, targets)
            elif task_type == "regression":
                loss_t = _rmse_torch(out, targets)
            else:
                raise ValueError("task_type must be 'classification' or 'regression'")

            fitness_val = float(loss_t.item()) + net_size_penalty * len(net)
            losses[i] = fitness_val

            if fitness_cache is not None:
                if sig not in fitness_cache:
                    fitness_cache[sig] = {}
                fitness_cache[sig]["fitness"] = None
                fitness_cache[sig]["weights"] = net._weights.detach().cpu().clone()

    return losses


def genotype_to_phenotype_tree(
    tree: Tree, n_variables: int, n_outputs: int, output_activation: str, offset: bool
) -> Net:
    pack: Any = []

    n = n_variables
    for node in reversed(tree._nodes):
        args = []
        for _ in range(node._n_args):
            args.append(pack.pop())
        if isinstance(node, FunctionalNode):
            pack.append(node._value(*args))
        else:
            if type(node) is TerminalNode:
                unit = Net(inputs=node._value)
            else:
                end = n + node._value._size
                hidden_id = set(range(n, end))
                activs = dict(zip(hidden_id, [node._value._activ] * len(hidden_id)))
                n = end
                unit = Net(hidden_layers=[hidden_id], activs=activs)
            pack.append(unit)
    end = n + n_outputs
    output_id = set(range(n, end))
    activs = dict(zip(output_id, [ACTIV_NAME_INV[output_activation]] * len(output_id)))
    to_return = pack[0] > Net(outputs=output_id, activs=activs)
    to_return = to_return._fix(set(range(n_variables)))
    to_return._offset = offset

    return to_return


def genotype_to_phenotype(
    population_g: NDArray,
    n_outputs: int,
    X_train: NDArray[np.float64],
    y_train: NDArray[np.float64],
    weights_optimizer_args: Dict,
    weights_optimizer_class: weights_type_optimizer_alias,
    output_activation: str,
    offset: bool,
    task_type: str = "regression",
    fitness_cache: Optional[dict[str, dict]] = None,
) -> NDArray:

    n_variables: int = X_train.shape[1]

    population_ph = np.array(
        [
            genotype_to_phenotype_tree(individ_g, n_variables, n_outputs, output_activation, offset)
            for individ_g in population_g
        ],
        dtype=object,
    )

    for i, net in enumerate(population_ph):
        sig = net.signature()

        if fitness_cache is not None and sig in fitness_cache:
            if fitness_cache[sig]["weights"] is not None:
                net._weights = fitness_cache[sig]["weights"].clone().to(X_train.device)
                net.compile_torch()
            continue

        population_ph[i]._weights, _ = train_net_weights(
            net=net,
            X_train=X_train,
            y_train=y_train,
            weights_optimizer_args=weights_optimizer_args,
            weights_optimizer=weights_optimizer_class,
            fitness_function=fitness_function_weights,
            task_type=task_type,
        )
        population_ph[i].compile_torch()

        if fitness_cache is not None:
            fitness_cache[sig] = {
                "fitness": None,
                "weights": net._weights.detach().cpu().clone(),
            }

    return population_ph


def train_net_structure(
    uniset: UniversalSet,
    X_train: NDArray[np.float64],
    y_train: NDArray[np.float64],
    X_test: NDArray[np.float64],
    y_test: NDArray[np.float64],
    optimizer_args: Dict,
    optimizer,
    fitness_function_structure: Callable,
    net_size_penalty: float,
    weights_optimizer_args: Dict,
    weights_optimizer: weights_type_optimizer_alias,
    offset: bool,
    n_outputs: int,
    task_type: str = "regression",
    fitness_cache: Optional[dict[str, dict]] = None,
):
    if task_type == "regression":
        output_activation = "ln"
    else:
        output_activation = "softmax"

    optimizer_args["fitness_function"] = fitness_function_structure
    optimizer_args["fitness_function_args"] = {
        "X": X_test,
        "targets": y_test,
        "net_size_penalty": net_size_penalty,
        "task_type": task_type,
        "fitness_cache": fitness_cache,
    }

    optimizer_args["genotype_to_phenotype"] = genotype_to_phenotype
    optimizer_args["genotype_to_phenotype_args"] = {
        "n_outputs": n_outputs,
        "X_train": X_train,
        "y_train": y_train,
        "weights_optimizer_args": weights_optimizer_args,
        "weights_optimizer_class": weights_optimizer,
        "output_activation": output_activation,
        "offset": offset,
        "task_type": task_type,
        "fitness_cache": fitness_cache,
    }

    optimizer_args["uniset"] = uniset
    optimizer_args["minimization"] = True

    optimizer = optimizer(**optimizer_args)

    optimizer.fit()

    net = optimizer.get_fittest()["phenotype"].copy()
    tree = optimizer.get_fittest()["genotype"].copy()

    return net, tree, optimizer._stats


class BaseGPNN(BaseEstimator, metaclass=ABCMeta):
    @abstractmethod
    def __init__(
        self,
        *,
        n_iter: int,
        pop_size: int,
        input_block_size: int = 1,
        max_hidden_block_size: int = 9,
        offset: bool = True,
        test_sample_ratio: float = 0.5,
        optimizer: Union[Type[SelfCGP], Type[GeneticProgramming]] = SelfCGP,
        optimizer_args: Optional[dict[str, Any]] = None,
        weights_optimizer: weights_type_optimizer_alias = SHAGA,
        weights_optimizer_args: Optional[dict[str, Any]] = None,
        net_size_penalty: float = 0.0,
        random_state: Optional[Union[int, np.random.RandomState]] = None,
        device: str = "cpu",
        use_fitness_cache: bool = False,
        fitness_cache_size: int = 1000,
    ):
        super().__init__()
        self.n_iter = n_iter
        self.pop_size = pop_size
        self.input_block_size = input_block_size
        self.max_hidden_block_size = max_hidden_block_size
        self.offset = offset
        self.test_sample_ratio = test_sample_ratio
        self.optimizer = optimizer
        self.optimizer_args = optimizer_args
        self.weights_optimizer = weights_optimizer
        self.weights_optimizer_args = weights_optimizer_args
        self.net_size_penalty = net_size_penalty
        self.random_state = random_state
        self.device = device
        self.use_fitness_cache = use_fitness_cache
        self.fitness_cache_size = fitness_cache_size

    def get_net(self) -> Net:
        return self.net_

    def get_tree(self) -> Tree:
        return self.tree_

    def get_stats(self) -> Statistics:
        return self.optimizer_stats_

    @staticmethod
    def genotype_to_phenotype_tree(
        tree: Tree, n_variables: int, n_outputs: int, output_activation: str, offset: bool
    ) -> Net:
        pack: Any = []

        n = n_variables
        for node in reversed(tree._nodes):
            args = []
            for _ in range(node._n_args):
                args.append(pack.pop())
            if isinstance(node, FunctionalNode):
                pack.append(node._value(*args))
            else:
                if type(node) is TerminalNode:
                    unit = Net(inputs=node._value)
                else:
                    end = n + node._value._size
                    hidden_id = set(range(n, end))
                    activs = dict(zip(hidden_id, [node._value._activ] * len(hidden_id)))
                    n = end
                    unit = Net(hidden_layers=[hidden_id], activs=activs)
                pack.append(unit)
        end = n + n_outputs
        output_id = set(range(n, end))
        activs = dict(zip(output_id, [ACTIV_NAME_INV[output_activation]] * len(output_id)))
        to_return = pack[0] > Net(outputs=output_id, activs=activs)
        to_return = to_return._fix(set(range(n_variables)))
        to_return._offset = offset

        return to_return

    def fit(self, X: ArrayLike, y: ArrayLike):

        if self.use_fitness_cache:
            self._fitness_cache = LimitedSizeDict(max_size=self.fitness_cache_size)
        else:
            self._fitness_cache = None

        if self.optimizer_args is not None:
            optimizer_args = self.optimizer_args.copy()
            check_optimizer_args(
                optimizer_args,
                args_auto_defined=[
                    "fitness_function",
                    "fitness_function_args",
                    "uniset",
                    "genotype_to_phenotype",
                    "genotype_to_phenotype_args",
                    "init_population",
                    "minimization",
                ],
                args_in_class=["iters", "pop_size"],
            )
        else:
            optimizer_args = {}
        optimizer_args["iters"] = self.n_iter
        optimizer_args["pop_size"] = self.pop_size

        if self.weights_optimizer_args is not None:
            weights_optimizer_args = self.weights_optimizer_args.copy()
            check_optimizer_args(
                weights_optimizer_args,
                args_auto_defined=[
                    "fitness_function",
                    "fitness_function_args",
                    "left_border",
                    "right_border",
                    "num_variables",
                    "str_len",
                    "genotype_to_phenotype",
                    "genotype_to_phenotype_args",
                    "minimization",
                    "init_population",
                ],
            )
        else:
            weights_optimizer_args = {}

        if "iters" not in weights_optimizer_args:
            weights_optimizer_args["iters"] = 300
        if "pop_size" not in weights_optimizer_args:
            weights_optimizer_args["pop_size"] = 300

        random_state = check_random_state(self.random_state)

        if isinstance(self, ClassifierMixin):
            if hasattr(self, "_validate_data"):
                X, y = self._validate_data(X, y, reset=True)
            else:
                X, y = validate_data(self, X, y, reset=True)

            check_classification_targets(y)
            self._label_encoder = LabelEncoder()
            self._one_hot_encoder = OneHotEncoder(
                sparse_output=False, categories="auto", dtype=np.float64
            )

            numeric_labels = self._label_encoder.fit_transform(y)
            y = self._one_hot_encoder.fit_transform(np.array(numeric_labels).reshape(-1, 1))
            self.classes_ = self._label_encoder.classes_
            self.n_classes_ = len(self.classes_)
        else:
            if hasattr(self, "_validate_data"):
                X, y = self._validate_data(X, y, reset=True)
            else:
                X, y = validate_data(self, X, y, reset=True)

        X, y = array_like_to_numpy_X_y(X, y)

        if self.offset:
            X = np.hstack([X, np.ones((X.shape[0], 1))])

        device = torch.device(self.device)
        X_t = torch.as_tensor(X, dtype=torch.float64, device=device)
        y_t = torch.as_tensor(y, dtype=torch.float64, device=device)

        X_train, X_test, y_train, y_test = train_test_split(
            X_t, y_t, test_size=self.test_sample_ratio, random_state=random_state
        )

        uniset: UniversalSet = init_net_uniset(
            n_variables=X.shape[1],
            input_block_size=self.input_block_size,
            max_hidden_block_size=self.max_hidden_block_size,
            offset=self.offset,
        )

        if isinstance(self, ClassifierMixin):
            task_type = "classification"
            n_outputs = self.n_classes_
        else:
            task_type = "regression"
            n_outputs = 1

        self.net_, self.tree_, self.optimizer_stats_ = train_net_structure(
            uniset=uniset,
            X_train=X_train,
            y_train=y_train,
            X_test=X_test,
            y_test=y_test,
            optimizer_args=optimizer_args,
            optimizer=self.optimizer,
            fitness_function_structure=fitness_function_structure,
            net_size_penalty=self.net_size_penalty,
            weights_optimizer_args=weights_optimizer_args,
            weights_optimizer=self.weights_optimizer,
            offset=self.offset,
            n_outputs=n_outputs,
            task_type=task_type,
            fitness_cache=self._fitness_cache,
        )
        self.net_.compile_torch()
        return self



import pytest

from sklearn.utils.estimator_checks import check_estimator

from ..benchmarks import BanknoteDataset
from ..benchmarks import IrisDataset
from ..classifiers import GeneticProgrammingClassifier
from ..classifiers import GeneticProgrammingNeuralNetClassifier
from ..classifiers import MLPEAClassifier
from ..optimizers import DifferentialEvolution
from ..optimizers import GeneticProgramming
from ..optimizers import SelfCGA
from ..optimizers import SelfCGP
from ..utils.transformations import minmax_scale


def test_GeneticProgrammingNeuralNetClassifier():
    data = IrisDataset()
    X = minmax_scale(data.get_X())
    y = data.get_y()

    model = GeneticProgrammingNeuralNetClassifier(
        n_iter=3,
        pop_size=10,
        optimizer=GeneticProgramming,
        optimizer_args={
            "tour_size": 15,
            "show_progress_each": 1,
            "mutation_rate": 0.07,
            "parents_num": 1,
            "elitism": False,
        },
        weights_optimizer=DifferentialEvolution,
        weights_optimizer_args={"iters": 25, "pop_size": 25, "CR": 0.9},
        max_hidden_block_size=2,
        offset=False,
        test_sample_ratio=0.3,
        net_size_penalty=0.01,
    )

    model.fit(X, y)
    model.predict(X)

    net = model.get_net()
    stats = model.get_stats()
    tree = model.get_tree()

    model = GeneticProgrammingNeuralNetClassifier(
        n_iter=3,
        pop_size=10,
        optimizer=GeneticProgramming,
        optimizer_args={
            "iters": 15,
            "tour_size": 15,
            "show_progress_each": 1,
            "mutation_rate": 0.07,
            "parents_num": 1,
            "elitism": False,
        },
        weights_optimizer=DifferentialEvolution,
        weights_optimizer_args={"iters": 25, "pop_size": 25, "CR": 0.9},
    )

    with pytest.raises(AssertionError):
        model.fit(X, y)

    model = GeneticProgrammingNeuralNetClassifier(
        n_iter=3,
        pop_size=10,
        optimizer=GeneticProgramming,
        optimizer_args={
            "tour_size": 15,
            "show_progress_each": 1,
            "mutation_rate": 0.07,
            "parents_num": 1,
            "elitism": False,
        },
        weights_optimizer=DifferentialEvolution,
        weights_optimizer_args={"iters": 25, "pop_size": 25, "CR": 0.9, "fitness_function": sum},
    )

    with pytest.raises(AssertionError):
        model.fit(X, y)

    model = GeneticProgrammingNeuralNetClassifier(
        n_iter=1, pop_size=10, weights_optimizer_args={"iters": 25, "pop_size": 25}
    )

    check_estimator(model)


def test_MLPEAClassifier():
    data = IrisDataset()
    X = minmax_scale(data.get_X())
    y = data.get_y()

    model = MLPEAClassifier(
        n_iter=15,
        pop_size=15,
        hidden_layers=(5, 3, 25),
        weights_optimizer=SelfCGA,
        weights_optimizer_args={"K": 0.9},
        activation="relu",
        offset=False,
    )

    model.fit(X, y)
    model.predict(X)

    net = model.get_net()
    stat = model.get_stats()

    model = MLPEAClassifier(
        n_iter=15,
        pop_size=15,
        hidden_layers=(5, 3, 25),
        weights_optimizer=SelfCGA,
        weights_optimizer_args={"iters": 100, "CR": 0.9},
        activation="relu",
        offset=False,
    )

    with pytest.raises(AssertionError):
        model.fit(X, y)

    model = MLPEAClassifier(hidden_layers=(0,), weights_optimizer_args={"no_increase_num": 30})
    check_estimator(model)

def test_GPClassifier():
    data = BanknoteDataset()
    X = minmax_scale(data.get_X())
    y = data.get_y()

    model = GeneticProgrammingClassifier(
        n_iter=15,
        pop_size=15,
        functional_set_names=("cos", "sin"),
        optimizer=SelfCGP,
        optimizer_args={"keep_history": True, "show_progress_each": 10, "elitism": True, "K": 2.5},
    )

    model.fit(X, y)
    model.predict(X)

    model = GeneticProgrammingClassifier(
        n_iter=15,
        pop_size=15,
        functional_set_names=("cos", "sin"),
        optimizer=SelfCGP,
        optimizer_args={
            "pop_size": 100,
            "keep_history": True,
            "show_progress_each": 10,
            "elitism": True,
            "K": 2.5,
        },
    )

    with pytest.raises(AssertionError):
        model.fit(X, y)

    model = GeneticProgrammingClassifier(
        n_iter=15,
        pop_size=15,
        functional_set_names=("cos2", "sin"),
        optimizer=SelfCGP,
        optimizer_args={
            "keep_history": True,
            "show_progress_each": 10,
            "elitism": True,
            "K": 2.5,
        },
    )

    with pytest.raises(ValueError):
        model.fit(X, y)

    model = GeneticProgrammingClassifier(
        n_iter=100, pop_size=100, optimizer_args={"no_increase_num": 30}
    )

    check_estimator(model)



from __future__ import annotations

from typing import Any
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union

import numpy as np
from numpy.typing import NDArray

from sklearn.base import RegressorMixin
from sklearn.utils.validation import check_is_fitted
from sklearn.utils.validation import validate_data

from ..base._gp import BaseGP
from ..optimizers import GeneticProgramming
from ..optimizers import SelfCGP


class GeneticProgrammingRegressor(RegressorMixin, BaseGP):
    def __init__(
        self,
        *,
        n_iter: int = 300,
        pop_size: int = 1000,
        functional_set_names: Tuple[str, ...] = ("cos", "sin", "add", "sub", "mul", "div"),
        optimizer: Union[Type[SelfCGP], Type[GeneticProgramming]] = SelfCGP,
        optimizer_args: Optional[dict[str, Any]] = None,
        random_state: Optional[Union[int, np.random.RandomState]] = None,
        use_fitness_cache: bool = False,
    ):
        super().__init__(
            n_iter=n_iter,
            pop_size=pop_size,
            functional_set_names=functional_set_names,
            optimizer=optimizer,
            optimizer_args=optimizer_args,
            random_state=random_state,
            use_fitness_cache=use_fitness_cache
        )

    def predict(self, X: NDArray[np.float64]):
        check_is_fitted(self)

        if hasattr(self, "_validate_data"):
            X = self._validate_data(X, reset=False)
        else:
            X = validate_data(self, X, reset=False)

        n_features = X.shape[1]
        tree_for_predict = self.tree_.set_terminals(**{f"x{i}": X[:, i] for i in range(n_features)})
        y_predict = tree_for_predict() * np.ones(len(X))
        return y_predict



from __future__ import annotations

from typing import Union

from numba import float64
from numba import int64
from numba import int8
from numba import njit

import numpy as np
from numpy.typing import NDArray

from .random import random_sample
from .random import sattolo_shuffle
from .random import flip_coin
from .random import randint
from .random import uniform
from ..base import EphemeralNode
from ..base import FunctionalNode
from ..base import TerminalNode
from ..base import Tree
from ..base import UniversalSet


@njit(int8[:](int8[:], float64))
def flip_mutation(individual: NDArray[np.byte], proba: float) -> NDArray[np.byte]:
    """
    Perform flip mutation on an individual's binary representation for a genetic algorithm.

    Parameters
    ----------
    individual : NDArray[np.byte]
        A 1D array representing the binary values of an individual.
    proba : float
        The probability of flipping each binary value in the individual.

    Returns
    -------
    NDArray[np.byte]
        A 1D array representing the mutated individual after applying flip mutation.

    Notes
    -----
    Flip mutation randomly flips binary values in the individual with the given probability.

    Examples
    --------
    >>> from thefittest.utils.mutations import flip_mutation
    >>> import numpy as np
    >>>
    >>> # Example
    >>> original_individual = np.array([0, 1, 1, 0, 1], dtype=np.byte)
    >>> mutation_probability = 0.1
    >>> mutated_individual = flip_mutation(original_individual, mutation_probability)
    >>> print("Original Individual:", original_individual)
    Original Individual: [0 1 1 0 1]
    >>> print("Mutated Individual:", mutated_individual)
    Mutated Individual: ...
    """
    offspring = individual.copy()
    for i in range(offspring.size):
        if flip_coin(proba):
            offspring[i] = 1 - offspring[i]
    return offspring


@njit(float64[:](float64[:], float64[:], float64[:, :], float64))
def best_1(
    current_individual: NDArray[np.float64],
    best_individual: NDArray[np.float64],
    population: NDArray[np.float64],
    F: np.float64,
) -> NDArray[np.float64]:
    """
    Perform Best-1 mutation on an individual in a population.

    Parameters
    ----------
    current_individual : NDArray[np.float64]
        The current individual undergoing mutation. (Not used in this mutation)
    best_individual : NDArray[np.float64]
        The best individual in the population.
    population : NDArray[np.float64]
        The entire population of individuals.
    F : float
        The mutation scale factor.

    Returns
    -------
    NDArray[np.float64]
        A mutated individual based on the Best-1 mutation strategy.

    Notes
    -----
    Best-1 mutation combines the best individual in the population with the difference between two randomly chosen
    individuals, scaled by the mutation factor F.

    Examples
    --------
    >>> from thefittest.utils.mutations import best_1
    >>> import numpy as np
    >>>
    >>> # Example
    >>> current_individual = np.array([0.5, 0.3, 0.8], dtype=np.float64)
    >>> best_individual = np.array([0.6, 0.4, 0.7], dtype=np.float64)
    >>> population = np.array([[0.2, 0.1, 0.5], [0.8, 0.6, 0.9], [0.3, 0.7, 0.4]], dtype=np.float64)
    >>> mutation_scale_factor = 0.7
    >>> mutated_individual = best_1(current_individual, best_individual, population, mutation_scale_factor)
    >>> print("Current Individual:", current_individual)
    Current Individual: [0.5 0.3 0.8]
    >>> print("Mutated Individual:", mutated_individual)
    Mutated Individual: ...
    """
    size = np.int64(len(population))
    r1, r2 = random_sample(range_size=size, quantity=np.int64(2), replace=False)
    mutated_individual = best_individual + F * (population[r1] - population[r2])
    return mutated_individual


@njit(float64[:](float64[:], float64[:], float64[:, :], float64))
def rand_1(
    current_individual: NDArray[np.float64],
    best_individual: NDArray[np.float64],
    population: NDArray[np.float64],
    F: np.float64,
) -> NDArray[np.float64]:
    """
    Perform Best-1 mutation on an individual in a population for differential evolution.

    Parameters
    ----------
    current_individual : NDArray[np.float64]
        The current individual undergoing mutation. (Not used in this mutation)
    best_individual : NDArray[np.float64]
        The best individual in the population. (Not used in this mutation)
    population : NDArray[np.float64]
        The entire population of individuals.
    F : float
        The mutation scale factor.

    Returns
    -------
    NDArray[np.float64]
        A mutated individual based on the Rand-1 mutation strategy.

    Notes
    -----
    Rand-1 mutation combines three randomly chosen individuals from the population, and the mutation is applied by
    adding the scaled difference between two of them to the third.

    Examples
    --------
    >>> from thefittest.utils.mutations import rand_1
    >>> import numpy as np
    >>>
    >>> # Example
    >>> current_individual = np.array([0.5, 0.3, 0.8], dtype=np.float64)
    >>> best_individual = np.array([0.6, 0.4, 0.7], dtype=np.float64)
    >>> population = np.array([[0.2, 0.1, 0.5], [0.8, 0.6, 0.9], [0.3, 0.7, 0.4]], dtype=np.float64)
    >>> mutation_scale_factor = 0.7
    >>> mutated_individual = rand_1(current_individual, best_individual, population, mutation_scale_factor)
    >>> print("Current Individual:", current_individual)
    Current Individual: [0.5 0.3 0.8]
    >>> print("Mutated Individual:", mutated_individual)
    Mutated Individual: ...
    """
    size = np.int64(len(population))
    r1, r2, r3 = random_sample(range_size=size, quantity=np.int64(3), replace=False)
    mutated_individual = population[r3] + F * (population[r1] - population[r2])
    return mutated_individual


@njit(float64[:](float64[:], float64[:], float64[:, :], float64))
def rand_to_best1(
    current_individual: NDArray[np.float64],
    best_individual: NDArray[np.float64],
    population: NDArray[np.float64],
    F: np.float64,
) -> NDArray[np.float64]:
    """
    Perform rand-to-best-1 mutation on an individual in a population for differential evolution.

    Parameters
    ----------
    current_individual : NDArray[np.float64]
        The current individual undergoing mutation. (Not used in this mutation)
    best_individual : NDArray[np.float64]
        The best individual in the population.
    population : NDArray[np.float64]
        The entire population of individuals.
    F : float
        The mutation scale factor.

    Returns
    -------
    NDArray[np.float64]
        A mutated individual based on the rand-to-best-1 mutation strategy.

    Notes
    -----
    Rand-to-best-1 mutation combines one randomly chosen individual with a scaled difference between the best individual
    and two other randomly chosen individuals from the population.

    Examples
    --------
    >>> from thefittest.utils.mutations import rand_to_best1
    >>> import numpy as np
    >>>
    >>> # Example
    >>> current_individual = np.array([0.5, 0.3, 0.8], dtype=np.float64)
    >>> best_individual = np.array([0.6, 0.4, 0.7], dtype=np.float64)
    >>> population = np.array([[0.2, 0.1, 0.5], [0.8, 0.6, 0.9], [0.3, 0.7, 0.4]], dtype=np.float64)
    >>> mutation_scale_factor = 0.7
    >>> mutated_individual = rand_to_best1(current_individual, best_individual, population, mutation_scale_factor)
    >>> print("Current Individual:", current_individual)
    Current Individual: [0.5 0.3 0.8]
    >>> print("Mutated Individual:", mutated_individual)
    Mutated Individual: ...
    """
    size = np.int64(len(population))
    r1, r2, r3 = random_sample(range_size=size, quantity=np.int64(3), replace=False)
    mutated_individual = (
        population[r1]
        + F * (population[r1] - best_individual)
        + F * (population[r2] - population[r3])
    )
    return mutated_individual


@njit(float64[:](float64[:], float64[:], float64[:, :], float64))
def current_to_best_1(
    current_individual: NDArray[np.float64],
    best_individual: NDArray[np.float64],
    population: NDArray[np.float64],
    F: np.float64,
) -> NDArray[np.float64]:
    """
    Perform current-to-best-1 mutation on an individual in a population for differential evolution.

    Parameters
    ----------
    current_individual : NDArray[np.float64]
        The current individual undergoing mutation.
    best_individual : NDArray[np.float64]
        The best individual in the population.
    population : NDArray[np.float64]
        The entire population of individuals.
    F : float
        The mutation scale factor.

    Returns
    -------
    NDArray[np.float64]
        A mutated individual based on the current-to-best-1 mutation strategy.

    Notes
    -----
    Current-to-best-1 mutation combines the current individual with a scaled difference between the best individual and
    the difference of two other randomly chosen individuals from the population.

    Examples
    --------
    >>> from thefittest.utils.mutations import current_to_best_1
    >>> import numpy as np
    >>>
    >>> # Example
    >>> current_individual = np.array([0.5, 0.3, 0.8], dtype=np.float64)
    >>> best_individual = np.array([0.6, 0.4, 0.7], dtype=np.float64)
    >>> population = np.array([[0.2, 0.1, 0.5], [0.8, 0.6, 0.9], [0.3, 0.7, 0.4]], dtype=np.float64)
    >>> mutation_scale_factor = 0.7
    >>> mutated_individual = current_to_best_1(current_individual, best_individual, population, mutation_scale_factor)
    >>> print("Current Individual:", current_individual)
    Current Individual: [0.5 0.3 0.8]
    >>> print("Mutated Individual:", mutated_individual)
    Mutated Individual: ...
    """
    size = np.int64(len(population))
    r1, r2 = random_sample(range_size=size, quantity=np.int64(2), replace=False)
    mutated_individual = (
        current_individual
        + F * (best_individual - current_individual)
        + F * (population[r1] - population[r2])
    )
    return mutated_individual


@njit(float64[:](float64[:], float64[:], float64[:, :], float64))
def best_2(
    current_individual: NDArray[np.float64],
    best_individual: NDArray[np.float64],
    population: NDArray[np.float64],
    F: np.float64,
) -> NDArray[np.float64]:
    """
    Perform Best-2 mutation on an individual in a population for differential evolution.

    Parameters
    ----------
    current_individual : NDArray[np.float64]
        The current individual undergoing mutation. (Not used in this mutation)
    best_individual : NDArray[np.float64]
        The best individual in the population.
    population : NDArray[np.float64]
        The entire population of individuals.
    F : float
        The mutation scale factor.

    Returns
    -------
    NDArray[np.float64]
        A mutated individual based on the Best-2 mutation strategy.

    Notes
    -----
    Best-2 mutation combines the best individual in the population with the difference between two pairs of randomly
    chosen individuals, each pair contributing to the mutation with a scaled difference, scaled by the mutation factor F.

    Examples
    --------
    >>> from thefittest.utils.mutations import best_2
    >>> import numpy as np
    >>>
    >>> # Example
    >>> current_individual = np.array([0.5, 0.3, 0.8], dtype=np.float64)
    >>> best_individual = np.array([0.6, 0.4, 0.7], dtype=np.float64)
    >>> population = np.array([[0.2, 0.1, 0.5], [0.8, 0.6, 0.9], [0.3, 0.7, 0.4], [0.9, 0.5, 0.2]], dtype=np.float64)
    >>> mutation_scale_factor = 0.7
    >>> mutated_individual = best_2(current_individual, best_individual, population, mutation_scale_factor)
    >>> print("Current Individual:", current_individual)
    Current Individual: [0.5 0.3 0.8]
    >>> print("Mutated Individual:", mutated_individual)
    Mutated Individual: ...
    """
    size = np.int64(len(population))
    r1, r2, r3, r4 = random_sample(range_size=size, quantity=np.int64(4), replace=False)
    mutated_individual = (
        best_individual
        + F * (population[r1] - population[r2])
        + F * (population[r3] - population[r4])
    )
    return mutated_individual


@njit(float64[:](float64[:], float64[:], float64[:, :], float64))
def rand_2(
    current_individual: NDArray[np.float64],
    best_individual: NDArray[np.float64],
    population: NDArray[np.float64],
    F: np.float64,
) -> NDArray[np.float64]:
    """
    Perform Rand-2 mutation on an individual in a population for differential evolution.

    Parameters
    ----------
    current_individual : NDArray[np.float64]
        The current individual undergoing mutation. (Not used in this mutation)
    best_individual : NDArray[np.float64]
        The best individual in the population. (Not used in this mutation)
    population : NDArray[np.float64]
        The entire population of individuals.
    F : float
        The mutation scale factor.

    Returns
    -------
    NDArray[np.float64]
        A mutated individual based on the Rand-2 mutation strategy.

    Notes
    -----
    Rand-2 mutation combines one randomly chosen individual with the difference between two pairs of randomly
    chosen individuals, each pair contributing to the mutation with a scaled difference, scaled by the mutation factor F.

    Examples
    --------
    >>> from thefittest.utils.mutations import rand_2
    >>> import numpy as np
    >>>
    >>> # Example
    >>> current_individual = np.array([0.5, 0.3, 0.8], dtype=np.float64)
    >>> best_individual = np.array([0.6, 0.4, 0.7], dtype=np.float64)
    >>> population = np.array([[0.2, 0.1, 0.5], [0.8, 0.6, 0.9], [0.3, 0.7, 0.4], [0.9, 0.5, 0.2], [0.4, 0.2, 0.6]], dtype=np.float64)
    >>> mutation_scale_factor = 0.7
    >>> mutated_individual = rand_2(current_individual, best_individual, population, mutation_scale_factor)
    >>> print("Current Individual:", current_individual)
    Current Individual: [0.5 0.3 0.8]
    >>> print("Mutated Individual:", mutated_individual)
    Mutated Individual: ...
    """
    size = np.int64(len(population))
    r1, r2, r3, r4, r5 = random_sample(range_size=size, quantity=np.int64(5), replace=False)
    mutated_individual = (
        population[r5]
        + F * (population[r1] - population[r2])
        + F * (population[r3] - population[r4])
    )
    return mutated_individual


@njit(float64[:](float64[:], float64[:], float64[:, :], float64))
def current_to_rand_1(
    current_individual: NDArray[np.float64],
    best_individual: NDArray[np.float64],
    population: NDArray[np.float64],
    F: np.float64,
) -> NDArray[np.float64]:
    """
    Perform current-to-rand-1 mutation on an individual in a population for differential evolution.

    Parameters
    ----------
    current_individual : NDArray[np.float64]
        The current individual undergoing mutation.
    best_individual : NDArray[np.float64]
        The best individual in the population. (Not used in this mutation)
    population : NDArray[np.float64]
        The entire population of individuals.
    F : float
        The mutation scale factor.

    Returns
    -------
    NDArray[np.float64]
        A mutated individual based on the current-to-rand-1 mutation strategy.

    Notes
    -----
    Current-to-rand-1 mutation combines the current individual with a scaled difference between one randomly chosen
    individual and the difference of two other randomly chosen individuals from the population.

    Examples
    --------
    >>> from thefittest.utils.mutations import current_to_rand_1
    >>> import numpy as np
    >>> # Example
    >>> current_individual = np.array([0.5, 0.3, 0.8], dtype=np.float64)
    >>> best_individual = np.array([0.6, 0.4, 0.7], dtype=np.float64)
    >>> population = np.array([[0.2, 0.1, 0.5], [0.8, 0.6, 0.9], [0.3, 0.7, 0.4]], dtype=np.float64)
    >>> mutation_scale_factor = 0.7
    >>> mutated_individual = current_to_rand_1(current_individual, best_individual, population, mutation_scale_factor)
    >>> print("Current Individual:", current_individual)
    Current Individual: [0.5 0.3 0.8]
    >>> print("Mutated Individual:", mutated_individual)
    Mutated Individual: ...
    """
    size = np.int64(len(population))
    r1, r2, r3 = random_sample(range_size=size, quantity=np.int64(3), replace=False)
    mutated_individual = (
        population[r1]
        + F * (population[r3] - current_individual)
        + F * (population[r1] - population[r2])
    )
    return mutated_individual


@njit(float64[:](float64[:], float64[:, :], int64[:], float64, float64[:, :]))
def current_to_pbest_1_archive(
    current: NDArray[np.float64],
    population: NDArray[np.float64],
    pbest: NDArray[np.int64],
    F: np.float64,
    pop_archive: NDArray[np.float64],
) -> NDArray[np.float64]:
    """
    Perform current-to-pbest-1-archive mutation on an individual in a population for differential evolution.

    Parameters
    ----------
    current : NDArray[np.float64]
        The current individual undergoing mutation.
    population : NDArray[np.float64]
        The entire population of individuals.
    pbest : NDArray[np.int64]
        Indices of the best individuals in the population.
    F : float
        The mutation scale factor.
    pop_archive : NDArray[np.float64]
        The archive of individuals from previous generations.

    Returns
    -------
    NDArray[np.float64]
        A mutated individual based on the current-to-pbest-1-archive mutation strategy.

    Notes
    -----
    Current-to-pbest-1-archive mutation combines the current individual with a scaled difference between one of the best
    individuals and the difference of two other randomly chosen individuals, one from the current population and one
    from the population archive.

    Examples
    --------
    >>> from thefittest.utils.mutations import current_to_pbest_1_archive
    >>> import numpy as np
    >>> # Example
    >>> current_individual = np.array([0.5, 0.3, 0.8], dtype=np.float64)
    >>> population = np.array([[0.2, 0.1, 0.5], [0.8, 0.6, 0.9], [0.3, 0.7, 0.4], [0.9, 0.5, 0.2]], dtype=np.float64)
    >>> pbest_indices = np.array([1, 3], dtype=np.int64)
    >>> mutation_scale_factor = 0.7
    >>> pop_archive = np.array([[0.1, 0.3, 0.6], [0.7, 0.4, 0.8], [0.4, 0.2, 0.5]], dtype=np.float64)
    >>> mutated_individual = current_to_pbest_1_archive(current_individual, population, pbest_indices, mutation_scale_factor, pop_archive)
    >>> print("Current Individual:", current_individual)
    Current Individual: [0.5 0.3 0.8]
    >>> print("Mutated Individual:", mutated_individual)
    Mutated Individual: ...
    """
    p_best_ind = randint(0, len(pbest), 1)[0]
    best = population[pbest[p_best_ind]]
    r1 = random_sample(range_size=len(population), quantity=np.int64(1), replace=True)[0]
    r2 = random_sample(range_size=len(pop_archive), quantity=np.int64(1), replace=True)[0]

    mutated_individual = current + F * (best - current) + F * (population[r1] - pop_archive[r2])
    return mutated_individual


@njit(float64[:](float64[:], float64[:, :], int64[:], float64, float64[:, :]))
def current_to_pbest_1_archive_p_min(
    current: NDArray[np.float64],
    population: NDArray[np.float64],
    pbest: NDArray[np.int64],
    F: np.float64,
    pop_archive: NDArray[np.float64],
) -> NDArray[np.float64]:
    """
    Perform current-to-pbest-1-archive mutation with a dynamically adjusted p_min on an individual in a population for differential evolution.

    Parameters
    ----------
    current : NDArray[np.float64]
        The current individual undergoing mutation.
    population : NDArray[np.float64]
        The entire population of individuals.
    pbest : NDArray[np.int64]
        Indices of the best individuals in the population.
    F : float
        The mutation scale factor.
    pop_archive : NDArray[np.float64]
        The archive of individuals from previous generations.

    Returns
    -------
    NDArray[np.float64]
        A mutated individual based on the current-to-pbest-1-archive mutation strategy with dynamically adjusted p_min.

    Notes
    -----
    Current-to-pbest-1-archive mutation with dynamically adjusted p_min combines the current individual with a scaled
    difference between one of the best individuals and the difference of two other randomly chosen individuals, one from
    the current population and one from the population archive. The parameter p_min is dynamically adjusted based on
    the size of the population.

    Examples
    --------
    >>> from thefittest.utils.mutations import current_to_pbest_1_archive_p_min
    >>> import numpy as np
    >>> # Example
    >>> current_individual = np.array([0.5, 0.3, 0.8], dtype=np.float64)
    >>> population = np.array([[0.2, 0.1, 0.5], [0.8, 0.6, 0.9], [0.3, 0.7, 0.4], [0.9, 0.5, 0.2]], dtype=np.float64)
    >>> pbest_indices = np.array([1, 3], dtype=np.int64)
    >>> mutation_scale_factor = 0.7
    >>> pop_archive = np.array([[0.1, 0.3, 0.6], [0.7, 0.4, 0.8], [0.4, 0.2, 0.5]], dtype=np.float64)
    >>> mutated_individual = current_to_pbest_1_archive_p_min(current_individual, population, pbest_indices, mutation_scale_factor, pop_archive)
    >>> print("Current Individual:", current_individual)
    Current Individual: [0.5 0.3 0.8]
    >>> print("Mutated Individual:", mutated_individual)
    Mutated Individual: ...
    """
    size = len(population)
    p_min = 2 / size
    p_i = uniform(p_min, 0.2, 1)[0]
    value = np.int64(max(1, p_i * size))
    pbest_cut = pbest[:value]

    p_best_ind = randint(0, len(pbest_cut), 1)[0]
    best = population[pbest_cut[p_best_ind]]
    r1 = random_sample(range_size=size, quantity=np.int64(1), replace=True)[0]
    r2 = random_sample(range_size=len(pop_archive), quantity=np.int64(1), replace=True)[0]

    mutated_individual = current + F * (best - current) + F * (population[r1] - pop_archive[r2])
    return mutated_individual


def point_mutation(tree: Tree, uniset: UniversalSet, proba: float, max_level: int) -> Tree:
    """
    Perform point mutation for a tree in genetic programming.

    Parameters
    ----------
    tree : Tree
        The input tree to undergo mutation.
    uniset : UniversalSet
        The universal set containing functional and terminal nodes for mutation.
    proba : float
        The probability of performing a point mutation at each node.
    max_level : int
        The maximum depth/level of the tree. (Unused in this mutation)

    Returns
    -------
    Tree
        A mutated tree after applying point mutation.

    Notes
    -----
    Point mutation randomly selects a node in the tree, and with the given probability, replaces it with a new node
    randomly chosen from the universal set.

    Examples
    --------
    >>> import numpy as np
    >>> from thefittest.base import Tree
    >>> from thefittest.base import init_symbolic_regression_uniset
    >>> from thefittest.utils.mutations import point_mutation
    >>> # Example Parameters
    >>> X = np.array([[0.3, 0.7], [0.3, 1.1], [3.5, 11.0]], dtype=np.float64)
    >>> functional_set_names = ("add", "mul")
    >>> max_tree_level = 4
    >>> mutation_probability = 0.5
    >>> # Initialize Universal Set for Symbolic Regression
    >>> universal_set = init_symbolic_regression_uniset(X, functional_set_names)
    >>> # Generate a Random Tree
    >>> tree = Tree.random_tree(universal_set, max_tree_level)
    >>> # Perform Point Mutation
    >>> mutated_tree = point_mutation(tree, universal_set, mutation_probability, max_tree_level)
    >>> print("Original Tree:", tree)
    Original Tree: ...
    >>> print("Mutated Tree:", mutated_tree)
    Mutated Tree: ...
    """
    new_node: Union[FunctionalNode, TerminalNode, EphemeralNode]

    mutated_tree = tree.copy()
    if flip_coin(proba):
        i = randint(0, len(mutated_tree), 1)[0]
        if isinstance(mutated_tree._nodes[i], FunctionalNode):
            n_args = mutated_tree._nodes[i]._n_args
            new_node = uniset._random_functional(n_args)
        else:
            new_node = uniset._random_terminal_or_ephemeral()
        mutated_tree._nodes[i] = new_node
    return mutated_tree


def growing_mutation(tree: Tree, uniset: UniversalSet, proba: float, max_level: int) -> Tree:
    """
    Perform growing mutation for a tree in genetic programming.

    Parameters
    ----------
    tree : Tree
        The input tree to undergo mutation.
    uniset : UniversalSet
        The universal set containing functional and terminal nodes for mutation.
    proba : float
        The probability of performing a growing mutation at each node.
    max_level : int
        The maximum depth/level of the tree. (Unused in this mutation)

    Returns
    -------
    Tree
        A mutated tree after applying growing mutation.

    Notes
    -----
    Growing mutation randomly selects a node in the tree, and with the given probability, replaces it with a new subtree
    generated by the growing method from the universal set. The growing method creates a tree with a random structure.

    Examples
    --------
    >>> import numpy as np
    >>> from thefittest.base import Tree
    >>> from thefittest.base import init_symbolic_regression_uniset
    >>> from thefittest.utils.mutations import growing_mutation
    >>> # Example Parameters
    >>> X = np.array([[0.3, 0.7], [0.3, 1.1], [3.5, 11.0]], dtype=np.float64)
    >>> functional_set_names = ("add", "mul")
    >>> max_tree_level = 4
    >>> mutation_probability = 0.5
    >>> # Initialize Universal Set for Symbolic Regression
    >>> universal_set = init_symbolic_regression_uniset(X, functional_set_names)
    >>> # Generate a Random Tree
    >>> tree = Tree.random_tree(universal_set, max_tree_level)
    >>> # Perform Growing Mutation
    >>> mutated_tree = growing_mutation(tree, universal_set, mutation_probability, max_tree_level)
    >>> print("Original Tree:", tree)
    Original Tree: ...
    >>> print("Mutated Tree:", mutated_tree)
    Mutated Tree: ...
    """
    mutated_tree = tree.copy()
    if flip_coin(proba):
        i = randint(0, len(mutated_tree), 1)[0]
        grown_tree = Tree.growing_method(uniset, max(mutated_tree.get_levels(i)))
        mutated_tree = mutated_tree.concat(i, grown_tree)
    return mutated_tree


def swap_mutation(tree: Tree, uniset: UniversalSet, proba: float, max_leve: int) -> Tree:
    """
    Perform swap mutation for a tree in genetic programming.

    Parameters
    ----------
    tree : Tree
        The input tree to undergo mutation.
    uniset : UniversalSet
        The universal set containing functional and terminal nodes for mutation. (Unused in this mutation)
    proba : float
        The probability of performing a swap mutation at each node.
    max_level : int
        The maximum depth/level of the tree. (Unused in this mutation)

    Returns
    -------
    Tree
        A mutated tree after applying swap mutation.

    Notes
    -----
    Swap mutation randomly selects a node in the tree, and with the given probability, swaps the positions of its
    subtrees, effectively changing the structure of the tree.

    Examples
    --------
    >>> import numpy as np
    >>> from thefittest.base import Tree
    >>> from thefittest.base import init_symbolic_regression_uniset
    >>> from thefittest.utils.mutations import swap_mutation
    >>> # Example Parameters
    >>> X = np.array([[0.3, 0.7], [0.3, 1.1], [3.5, 11.0]], dtype=np.float64)
    >>> functional_set_names = ("add", "mul")
    >>> max_tree_level = 4
    >>> mutation_probability = 0.5
    >>> # Initialize Universal Set for Symbolic Regression
    >>> universal_set = init_symbolic_regression_uniset(X, functional_set_names)
    >>> # Generate a Random Tree
    >>> tree = Tree.random_tree(universal_set, max_tree_level)
    >>> # Perform Swap Mutation
    >>> mutated_tree = swap_mutation(tree, universal_set, mutation_probability, max_tree_level)
    >>> print("Original Tree:", tree)
    Original Tree: ...
    >>> print("Mutated Tree:", mutated_tree)
    Mutated Tree: ...
    """
    mutated_tree = tree.copy()
    if flip_coin(proba):
        more_one_args_cond = mutated_tree._n_args > 1
        indexes = np.arange(len(tree), dtype=int)[more_one_args_cond]
        if len(indexes) > 0:
            index = random_sample(range_size=len(indexes), quantity=1, replace=True)[0]
            i = indexes[index]
            args_id = mutated_tree.get_args_id(i)
            new_arg_id = args_id.copy()
            new_arg_id = sattolo_shuffle(new_arg_id)
            for old_j, new_j in zip(args_id, new_arg_id):
                subtree = tree.subtree(old_j)
                mutated_tree = mutated_tree.concat(new_j, subtree)
    return mutated_tree


def shrink_mutation(tree: Tree, uniset: UniversalSet, proba: float, max_level: int) -> Tree:
    """
    Perform shrink mutation for a tree in genetic programming.

    Parameters
    ----------
    tree : Tree
        The input tree to undergo mutation.
    uniset : UniversalSet
        The universal set containing functional and terminal nodes for mutation. (Unused in this mutation)
    proba : float
        The probability of performing a shrink mutation at each node.
    max_level : int
        The maximum depth/level of the tree. (Unused in this mutation)

    Returns
    -------
    Tree
        A mutated tree after applying shrink mutation.

    Notes
    -----
    Shrink mutation randomly selects a node in the tree, and with the given probability, replaces it with one of its
    terminal nodes, effectively pruning the subtree rooted at the selected node.

    Examples
    --------
    >>> import numpy as np
    >>> from thefittest.base import Tree
    >>> from thefittest.base import init_symbolic_regression_uniset
    >>> from thefittest.utils.mutations import shrink_mutation
    >>> # Example Parameters
    >>> X = np.array([[0.3, 0.7], [0.3, 1.1], [3.5, 11.0]], dtype=np.float64)
    >>> functional_set_names = ("add", "mul")
    >>> max_tree_level = 4
    >>> mutation_probability = 0.5
    >>> # Initialize Universal Set for Symbolic Regression
    >>> universal_set = init_symbolic_regression_uniset(X, functional_set_names)
    >>> # Generate a Random Tree
    >>> tree = Tree.random_tree(universal_set, max_tree_level)
    >>> # Perform Shrink Mutation
    >>> mutated_tree = shrink_mutation(tree, universal_set, mutation_probability, max_tree_level)
    >>> print("Original Tree:", tree)
    Original Tree: ...
    >>> print("Mutated Tree:", mutated_tree)
    Mutated Tree: ...
    """
    mutated_tree = tree.copy()
    if len(mutated_tree) > 2:
        if flip_coin(proba):
            no_terminal_cond = mutated_tree._n_args > 0
            indexes = np.arange(len(tree), dtype=np.int64)[no_terminal_cond]
            if len(indexes) > 0:
                index = random_sample(range_size=len(indexes), quantity=1, replace=True)[0]
                i = indexes[index]

                args_id = mutated_tree.get_args_id(i)

                if len(args_id) > 1:
                    index = random_sample(range_size=len(args_id), quantity=1, replace=True)[0]
                    choosen_id = args_id[index]
                else:
                    choosen_id = args_id[0]

                mutated_tree = mutated_tree.concat(i, tree.subtree(choosen_id))
    return mutated_tree
